# -*- coding: utf-8 -*-

from imio.helpers import HAS_PLONE_4
from imio.helpers.cache import get_cachekey_volatile
from imio.helpers.cache import get_plone_groups_for_user
from imio.helpers.cache import invalidate_cachekey_volatile_for
from plone.api.exc import InvalidParameterError
from plone.memoize import ram
from Products.PlonePAS.plugins.role import GroupAwareRoleManager
from Products.PluggableAuthService.plugins.ZODBRoleManager import ZODBRoleManager
from smtplib import SMTP_SSL
from zope.sendmail.mailer import SMTPMailer

import pkg_resources


try:
    pkg_resources.get_distribution("collective.solr")
except pkg_resources.DistributionNotFound:
    HAS_SOLR = False
else:
    from collective.solr.indexer import SolrIndexProcessor

    HAS_SOLR = True


def solr_index(self, obj, attributes=None):
    # Fix issue https://github.com/collective/collective.solr/issues/189
    if attributes is not None:
        attributes = None
    return SolrIndexProcessor._old_index(self, obj, attributes)


def ssl_makeMailer(self):
    """Create a SMTPMailer with SMTP_SSL if port is 465"""
    mailer = SMTPMailer(
        hostname=self.smtp_host,
        port=int(self.smtp_port),
        username=self.smtp_uid or None,
        password=self.smtp_pwd or None,
        force_tls=self.force_tls,
    )
    if self.smtp_port in (465, "465"):
        mailer.smtp = SMTP_SSL
    return mailer


def assignRolesToPrincipal(self, roles, principal_id, REQUEST=None):  # noqa
    GroupAwareRoleManager._old_assignRolesToPrincipal(self, roles, principal_id, REQUEST)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for("_users_groups_value")


def assignRoleToPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_assignRoleToPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for("_users_groups_value")


def removeRoleFromPrincipal(self, role_id, principal_id):
    ZODBRoleManager._old_removeRoleFromPrincipal(self, role_id, principal_id)
    # we need to invalidate cachekey
    invalidate_cachekey_volatile_for("_users_groups_value")


def _listAllowedRolesAndUsers_cachekey(method, self, user):
    """cachekey method for self._listAllowedRolesAndUsers."""
    date = get_cachekey_volatile("_users_groups_value")
    return date, user.getId()


@ram.cache(_listAllowedRolesAndUsers_cachekey)
def _listAllowedRolesAndUsers(self, user):
    """Monkeypatch to use get_plone_groups_for_user instead getGroups.
    Moreover store this in the REQUEST."""
    # Makes sure the list includes the user's groups.
    result = user.getRoles()
    if "Anonymous" in result:
        # The anonymous user has no further roles
        return ["Anonymous"]
    result = list(result)
    # XXX change, replaced getGroups by get_plone_groups_for_user
    # if hasattr(aq_base(user), 'getGroups'):
    #     groups = ['user:%s' % x for x in user.getGroups()]
    try:
        groups = get_plone_groups_for_user(user=user)
    except InvalidParameterError:
        groups = user.getGroups()
    if groups:
        groups = ["user:%s" % x for x in groups]
        result = result + groups
    # Order the arguments from small to large sets
    result.insert(0, "user:%s" % user.getId())
    result.append("Anonymous")
    return result


if HAS_PLONE_4:
    from operator import attrgetter
    from plone.app.querystring.registryreader import logger
    from zope.component import queryUtility
    from zope.i18n import translate
    from zope.i18nmessageid import Message
    from zope.schema.interfaces import IVocabularyFactory

    def getVocabularyValues(self, values):
        """Get all vocabulary values if a vocabulary is defined"""

        for field in values.get(self.prefix + ".field").values():
            # XXX begin change by imio.helpers
            # field['values'] = {}
            from collections import OrderedDict

            field["values"] = OrderedDict()
            # XXX end change by imio.helpers
            vocabulary = field.get("vocabulary", [])
            if vocabulary:
                utility = queryUtility(IVocabularyFactory, vocabulary)
                if utility is not None:
                    for item in sorted(utility(self.context), key=attrgetter("title")):
                        if isinstance(item.title, Message):
                            title = translate(item.title, context=self.request)
                        else:
                            title = item.title

                        field["values"][item.value] = {"title": title}
                else:
                    logger.info("%s is missing, ignored." % vocabulary)
        return values



# -*- coding: utf-8 -*-

from datetime import datetime
from datetime import timedelta
from persistent.mapping import PersistentMapping
from plone import api
from plone.i18n.normalizer import IIDNormalizer
from plone.memoize import ram
from plone.memoize.forever import _memos
from plone.memoize.instance import Memojito
from plone.memoize.interfaces import ICacheChooser
from Products.CMFPlone.utils import base_hasattr
from types import FunctionType
from zope.component import getAllUtilitiesRegisteredFor
from zope.component import getUtility
from zope.component import queryUtility
from zope.globalrequest import getRequest
from zope.ramcache.interfaces.ram import IRAMCache
from zope.schema.interfaces import IVocabularyFactory

import logging


logger = logging.getLogger("imio.helpers:cache")
VOLATILE_NAME_MAX_LENGTH = 200
VOLATILE_ATTR = "_volatile_cache_keys"
METHODS_MAPPING_NAME = "_methods_invalidation_mapping"


def cleanVocabularyCacheFor(vocabulary=None):
    """Clean _memojito_ attribute for given p_vocabulary name.
    If p_vocabulary is None, it will clean every vocabulary _memojito_ attribute."""

    # we received a vocabulary name, just clean this one
    if vocabulary:
        vocabularies = (queryUtility(IVocabularyFactory, vocabulary),)
    else:
        # clean every vocabularies having a _memojito_ attribute
        vocabularies = getAllUtilitiesRegisteredFor(IVocabularyFactory)

    for vocab in vocabularies:
        if hasattr(vocab, Memojito.propname):
            getattr(vocab, Memojito.propname).clear()


def cleanRamCache():
    """Clean the entire ram.cache."""
    cache_chooser = getUtility(ICacheChooser)
    thecache = cache_chooser("")
    thecache.ramcache.invalidateAll()


def cleanRamCacheFor(methodId):
    """Clean ram.cache for given p_methodId."""
    cache_chooser = getUtility(ICacheChooser)
    thecache = cache_chooser(methodId)
    thecache.ramcache.invalidate(methodId)


def cleanForeverCache():
    """Clean cache using the @forever.memoize decorator."""
    _memos.clear()


def get_cachekey_volatile(name, method=None, ttl=0):
    """Helper for using a volatile corresponding to p_name
    to be used as cachekey stored in a volatile.
    If it exists, we return the value, either we store datetime.now().
    If p_ttl (time to live) is given, a cachekey older that ttl is updated."""
    portal = api.portal.get()
    # use max_length of VOLATILE_NAME_MAX_LENGTH to avoid cropped names
    # that could lead to having 2 names beginning with same part using same volatile...
    normalized_name = queryUtility(IIDNormalizer).normalize(name, max_length=VOLATILE_NAME_MAX_LENGTH)
    volatile_name = normalized_name
    volatiles = getattr(portal, VOLATILE_ATTR, None)
    if volatiles is None:
        portal._volatile_cache_keys = PersistentMapping()
        volatiles = portal._volatile_cache_keys
    date = volatiles.get(volatile_name)
    now = datetime.now()
    # compute new date if None or if using ttl and ttl is stale
    if not date or (ttl and date + timedelta(seconds=ttl) < now):
        date = now
        volatiles[volatile_name] = date
    # store caller method path so it will be invalidated in invalidate_cachekey_volatile_for
    if method:
        key = "%s.%s" % (method.__module__, method.__name__)
        methods = volatiles.get(METHODS_MAPPING_NAME)
        if methods is None:
            volatiles[METHODS_MAPPING_NAME] = PersistentMapping()
        if name not in volatiles[METHODS_MAPPING_NAME]:
            volatiles[METHODS_MAPPING_NAME][name] = []
        if key not in volatiles[METHODS_MAPPING_NAME][name]:
            volatiles[METHODS_MAPPING_NAME][name].append(key)
    return date


def invalidate_cachekey_volatile_for(name, get_again=False, invalidate_cache=True):
    """ """
    portal = api.portal.get()
    normalized_name = queryUtility(IIDNormalizer).normalize(name, max_length=VOLATILE_NAME_MAX_LENGTH)
    volatile_name = normalized_name
    volatiles = getattr(portal, VOLATILE_ATTR, {})
    if volatile_name in volatiles:
        del volatiles[volatile_name]
    # if get_again=True, when the key is invalidated,
    # get_cachekey_volatile so it stores a new date and it avoids a second write
    date = get_cachekey_volatile(volatile_name) if get_again else None
    # when date is invalidated, every cache using it is stale
    # so we may either specifically invalidate this cached methods
    # or just wait for ram.cache to do it's cleanup itself
    if invalidate_cache:
        mapping = volatiles.get(METHODS_MAPPING_NAME, {})
        if name in mapping:
            for method in mapping[name]:
                cleanRamCacheFor(method)
            mapping.pop(name)
    return date


def _generate_params_key(*args, **kwargs):
    items = []
    for item in kwargs.items():
        elements = []
        for i in item:
            if isinstance(i, list):
                i = tuple(i)
            elements.append(i)
        items.append(tuple(elements))
    return (args, frozenset(items))


def generate_key(func):
    """Return the complete path for a function e.g. module.function"""
    if hasattr(func, "_cache_key"):
        return func._cache_key
    path = [func.__module__]
    if hasattr(func, "im_class"):  # unbounded func (no more in py3)
        path.append(func.im_class.__name__)
        path.append(func.__name__)
    elif hasattr(func, "__qualname__"):
        path.append(func.__qualname__)
    else:
        path.append(func.__name__)
    return ".".join(path)


def volatile_cache_with_parameters(func):
    def get_key(func, *args, **kwargs):
        return (
            get_cachekey_volatile(generate_key(func)),
            _generate_params_key(*args, **kwargs),
        )

    def cache(get_key):
        return ram.cache(get_key)

    replacement = cache(get_key)(func)
    replacement._cache_key = generate_key(func)
    return replacement


def volatile_cache_without_parameters(func):
    def get_key(func, *args, **kwargs):
        return (
            get_cachekey_volatile(generate_key(func)),
            (),
        )

    def cache(get_key):
        return ram.cache(get_key)

    replacement = cache(get_key)(func)
    replacement._cache_key = generate_key(func)
    return replacement


def obj_modified(obj, asdatetime=True, check_annotation=True):
    """Returns max value between obj.modified(), obj._p_mtime and __anotations__._p_mtime
    to check also attribute modification and annotation modification.
    Returns a float when asdatetime=False.
    """
    modified = max(float(obj.modified()), obj._p_mtime or 0)
    if check_annotation and base_hasattr(obj, "__annotations__"):
        # when stored annotation is a PersistentMapping, we need to check _p_mtime
        # of stored annotation because a change in the PersistentMapping will not change
        # the __annotations__ _p_mtime
        ann_max_time = max(
            [obj.__annotations__._p_mtime or 0]
            + [
                obj.__annotations__[k]._p_mtime
                for k in obj.__annotations__.keys()
                if hasattr(obj.__annotations__[k], "_p_mtime") and obj.__annotations__[k]._p_mtime is not None
            ]
        )
        modified = max(float(obj.modified()), obj._p_mtime or 0, ann_max_time)
    if asdatetime:
        modified = datetime.fromtimestamp(modified)
    return modified


def extract_wrapped(decorated):
    """Returns function that was decorated"""
    closure = (c.cell_contents for c in decorated.__closure__)
    return next((c for c in closure if isinstance(c, FunctionType)), None)


def setup_ram_cache(max_entries=100000, max_age=2400, cleanup_interval=600):
    """Can be called in IProcessStarting subscriber"""
    ramcache = queryUtility(IRAMCache)
    logger.info(
        "=> Setting ramcache parameters (maxEntries=%s, maxAge=%s, cleanupInterval=%s)"
        % (max_entries, max_age, cleanup_interval)
    )
    ramcache.update(maxEntries=max_entries, maxAge=max_age, cleanupInterval=cleanup_interval)


def get_current_user_id(request=None):
    """Try to get user_id from REQUEST or fallback to plone.api."""
    user_id = None
    try:
        if request is None:
            request = getRequest()
        user_id = request["AUTHENTICATED_USER"].getId()
    except Exception:
        user_id = api.user.get_current().getId()
    return user_id


def get_plone_groups_for_user_cachekey(method, user_id=None, user=None, the_objects=False):
    """cachekey method for self.get_plone_groups_for_user."""
    date = get_cachekey_volatile("_users_groups_value")
    return (
        date,
        # use user.getId() and not user.id because when user is a PloneUser instance
        # it does not have a "id" and returns "acl_users" which break the invalidation
        user and user.getId() or user_id or get_current_user_id(getRequest()),
        the_objects,
    )


@ram.cache(get_plone_groups_for_user_cachekey)
def get_plone_groups_for_user(user_id=None, user=None, the_objects=False):
    """Just return user.getGroups but cached.
    This method is tested in
    Products.PloneMeeting.tests.testToolPloneMeeting.test_pm_Get_plone_groups_for_user."""
    if api.user.is_anonymous():
        return []
    if user is None:
        user = user_id and api.user.get(user_id) or api.user.get_current()
    if not hasattr(user, "getGroups"):
        return []
    if the_objects:
        pg = api.portal.get_tool("portal_groups")
        user_groups = pg.getGroupsByUserId(user.getId())
    else:
        user_groups = user.getGroups()
    return sorted(user_groups)


def get_users_in_plone_groups_cachekey(method, group_id=None, group=None, the_objects=False):
    """cachekey method for self.get_users_in_plone_groups."""
    date = get_cachekey_volatile("_users_groups_value")
    return (date, group and group.getId() or group_id, the_objects)


@ram.cache(get_users_in_plone_groups_cachekey)
def get_users_in_plone_groups(group_id=None, group=None, the_objects=False):
    """Just return group.getGroupMembers but cached."""
    if not group and group_id:
        group = api.group.get(group_id)
    if group is None:
        return []
    members = group.getGroupMembers()
    if not the_objects:
        members = [m.getId() for m in members]
    return sorted(members)



# -*- coding: utf-8 -*-

from imio.helpers import _
from imio.helpers import HAS_PLONE_6_AND_MORE
from plone import api
from plone.namedfile.file import NamedBlobImage
from zope.interface import Invalid

import json


def create_image_content(container, title, id, data, filename=None, portal_type="Image"):
    if HAS_PLONE_6_AND_MORE:
        obj = api.content.create(
            container=container,
            type=portal_type,
            id=id,
            title=title,
        )
        filename = filename or id
        obj.image = NamedBlobImage(**{"filename": filename, "data": data})
    else:
        obj = api.content.create(
            container=container,
            type=portal_type,
            id=id,
            title=title,
            file=data,
        )
    return obj


def is_valid_json(value):
    if value:
        try:
            json.loads(value)
        except:  # NOQA: E722
            raise Invalid(_(u"Invalid JSON."))
    return True



# -*- coding: utf-8 -*-
from imio.helpers.batching import batch_get_keys
from imio.helpers.batching import batch_handle_key
from imio.helpers.batching import batch_hashed_filename
from imio.helpers.batching import batch_loop_else
from imio.helpers.batching import batch_skip_key
from imio.helpers.batching import can_delete_batch_files
from imio.pyutils.batching import batch_delete_files

import logging
import os
import transaction
import unittest


KEYS_PKL_FILE = "keys.pkl"
CONFIG_FILE = "keys_config.txt"
processed = {"keys": [], "commits": 0, "errors": 0}


def loop_process(loop_len, batch_number, commit_number, a_set, last=False, add_files=()):
    """Process the loop using the batching module."""
    os.environ["BATCH"] = str(batch_number)
    os.environ["COMMIT"] = str(commit_number)
    os.environ["BATCH_LAST"] = str(int(last))
    batch_keys, config = batch_get_keys(KEYS_PKL_FILE, loop_len, a_set=a_set, add_files=add_files)
    for key in range(1, loop_len + 1):
        if batch_skip_key(key, batch_keys, config):
            continue
        processed["keys"].append(key)
        if batch_handle_key(key, batch_keys, config):
            break
    else:
        batch_loop_else(batch_keys, config)
    if can_delete_batch_files(batch_keys, config):
        batch_delete_files(batch_keys, config, rename=False)
    return batch_keys, config


def fake_transaction_commit():
    """Fake transaction commit."""
    processed["commits"] += 1


def fake_logger_error(msg, *args, **kwargs):
    """Fake logger error."""
    processed["errors"] += 1


def remove_files():
    for fil in (KEYS_PKL_FILE, CONFIG_FILE):
        if os.path.exists(fil):
            os.remove(fil)


def reset_processed():
    processed["keys"] = []
    processed["commits"] = 0
    processed["errors"] = 0


class TestBatching(unittest.TestCase):
    def setUp(self):
        super(TestBatching, self).setUp()
        remove_files()

    def test_batching(self):
        orig_tc_func = transaction.commit
        transaction.commit = fake_transaction_commit
        logger = logging.getLogger("imio.helpers")
        orig_le_func = logger.error
        logger.error = fake_logger_error
        a_set = set()
        # no items
        keys, conf = loop_process(0, 0, 0, a_set)  # ll, bn, cn
        self.assertEqual(processed["keys"], [])
        self.assertEqual(processed["commits"], 0)
        self.assertSetEqual(a_set, set())
        self.assertEqual(conf["kc"], 0)
        self.assertEqual(conf["lc"], 0)
        self.assertEqual(conf.get("lk"), None)
        self.assertListEqual(conf["af"], [])
        self.assertIsNone(keys)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # no batching
        keys, conf = loop_process(5, 0, 0, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 0)
        self.assertSetEqual(a_set, set())
        self.assertEqual(conf["kc"], 0)
        self.assertEqual(conf["lc"], 0)
        self.assertEqual(conf.get("lk"), None)
        self.assertIsNone(keys)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # no batching but commit used
        reset_processed()
        keys, conf = loop_process(5, 0, 5, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 0)
        self.assertSetEqual(a_set, set())
        self.assertEqual(conf["kc"], 0)
        self.assertEqual(conf["lc"], 0)
        self.assertEqual(conf.get("lk"), None)
        self.assertIsNone(keys)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # batching: 2 passes with commit each item
        reset_processed()
        with open("various.pkl", "w"):
            pass
        keys, conf = loop_process(5, 3, 1, a_set, add_files=["various.pkl"])
        self.assertEqual(processed["keys"], [1, 2, 3])
        self.assertEqual(processed["commits"], 3)
        self.assertSetEqual(a_set, {1, 2, 3})
        self.assertEqual(conf["kc"], 3)
        self.assertEqual(conf["lc"], 3)
        self.assertEqual(conf.get("lk"), 3)
        self.assertSetEqual(keys, a_set)
        self.assertTrue(os.path.exists(KEYS_PKL_FILE))
        self.assertTrue(os.path.exists(CONFIG_FILE))
        self.assertEqual(len(conf["af"]), 1)
        self.assertTrue(os.path.exists("various.pkl"))
        keys, conf = loop_process(5, 3, 1, a_set, last=True, add_files=["various.pkl"])
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 5)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 2)
        self.assertEqual(conf.get("lk"), 5)
        self.assertSetEqual(keys, a_set)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        self.assertFalse(os.path.exists("various.pkl"))
        # batching: 2 passes with commit each 3 items
        reset_processed()
        a_set = set()
        remove_files()
        keys, conf = loop_process(5, 3, 3, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3])
        self.assertEqual(processed["commits"], 1)
        self.assertSetEqual(a_set, {1, 2, 3})
        self.assertEqual(conf["kc"], 3)
        self.assertEqual(conf["lc"], 3)
        self.assertEqual(conf.get("lk"), 3)
        self.assertSetEqual(keys, a_set)
        self.assertTrue(os.path.exists(KEYS_PKL_FILE))
        self.assertTrue(os.path.exists(CONFIG_FILE))
        keys, conf = loop_process(5, 3, 3, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 2)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertSetEqual(keys, a_set)
        self.assertTrue(os.path.exists(KEYS_PKL_FILE))
        self.assertTrue(os.path.exists(CONFIG_FILE))
        keys, conf = loop_process(5, 3, 3, a_set, last=True)  # what if one more call ?
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 2)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 0)
        self.assertEqual(conf.get("lk"), None)
        self.assertSetEqual(keys, a_set)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # batching: 1 pass with commit each 3 items by imio.updates
        reset_processed()
        os.environ["IU_RUN1"] = "1"
        remove_files()
        a_set = set()
        keys, conf = loop_process(5, 10, 3, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 2)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 5)
        self.assertEqual(conf.get("lk"), 5)
        self.assertSetEqual(keys, a_set)
        self.assertTrue(os.path.exists(KEYS_PKL_FILE))
        self.assertTrue(os.path.exists(CONFIG_FILE))
        # batching: 1 pass with commit each 3 items
        reset_processed()
        del os.environ["IU_RUN1"]
        remove_files()
        a_set = set()
        keys, conf = loop_process(5, 10, 3, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 2)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 5)
        self.assertEqual(conf.get("lk"), 5)
        self.assertSetEqual(keys, a_set)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # batching: 1 pass with commit greather than items lentgh
        reset_processed()
        a_set = set()
        keys, conf = loop_process(5, 10, 10, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 1)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 5)
        self.assertEqual(conf.get("lk"), 5)
        self.assertSetEqual(keys, a_set)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # batching: 1 pass with 1 commit
        reset_processed()
        a_set = set()
        keys, conf = loop_process(5, 5, 5, a_set)
        self.assertEqual(processed["keys"], [1, 2, 3, 4, 5])
        self.assertEqual(processed["commits"], 1)
        self.assertSetEqual(a_set, {1, 2, 3, 4, 5})
        self.assertEqual(conf["kc"], 5)
        self.assertEqual(conf["lc"], 5)
        self.assertEqual(conf.get("lk"), 5)
        self.assertSetEqual(keys, a_set)
        self.assertFalse(os.path.exists(KEYS_PKL_FILE))
        self.assertFalse(os.path.exists(CONFIG_FILE))
        # batching: finishing too early
        reset_processed()
        a_set = set()
        keys, conf = loop_process(5, 3, 5, a_set, last=True)
        self.assertEqual(processed["keys"], [1, 2, 3])
        self.assertEqual(processed["commits"], 1)
        self.assertEqual(processed["errors"], 1)  # finished too early
        self.assertSetEqual(a_set, {1, 2, 3})
        self.assertEqual(conf["kc"], 3)
        self.assertEqual(conf["lc"], 3)
        self.assertEqual(conf.get("lk"), 3)
        self.assertSetEqual(keys, a_set)
        self.assertTrue(os.path.exists(KEYS_PKL_FILE))
        self.assertTrue(os.path.exists(CONFIG_FILE))

        transaction.commit = orig_tc_func
        logger.error = orig_le_func

    def test_batch_hashed_filename(self):
        nfn = batch_hashed_filename("file.txt")
        self.assertTrue(len(os.path.dirname(nfn)) > 1)
        self.assertEqual(os.path.basename(nfn), "file.txt")



# -*- coding: utf-8 -*-
from imio.helpers.content import get_vocab
from imio.helpers.testing import IntegrationTestCase
from plone import api
from zope.schema.tests.test_vocabulary import SimpleTermTests


class TestVocabularies(IntegrationTestCase):
    """
    Test vocabularies.
    """

    def test_SortedUsers(self):
        """ """
        api.user.create(email="test@test.org", username="a_new_user", password="secret1234")
        api.user.create(email="test@test.org", username="b_new_user", password="secret1234")
        api.user.create(email="test@test.org", username="new_user", password="secret1234")
        api.user.create(email="test@test.org", username="A_new_user", password="secret1234")
        api.user.create(email="test@test.org", username="B_new_user", password="secret1234")
        vocab = get_vocab(self.portal, "imio.helpers.SortedUsers")
        self.assertEqual(
            [term.title for term in vocab._terms],
            [
                u"a_new_user (a_new_user)",
                u"A_new_user (A_new_user)",
                u"b_new_user (b_new_user)",
                u"B_new_user (B_new_user)",
                u"new_user (new_user)",
                u"test_user_1_ (test_user_1_)",
            ],
        )
        vocab2 = get_vocab(self.portal, "imio.helpers.SimplySortedUsers")
        self.assertEqual(
            [term.title for term in vocab2._terms],
            [u"a_new_user", u"A_new_user", u"b_new_user", u"B_new_user", u"new_user", u"test_user_1_"],
        )
        member = api.user.get(username="a_new_user")
        member.setMemberProperties({"fullname": "anew User"})
        vocab = get_vocab(self.portal, "imio.helpers.SortedUsers")
        vocab2 = get_vocab(self.portal, "imio.helpers.SimplySortedUsers")
        self.assertEqual(
            [term.title for term in vocab._terms],
            [
                u"anew User (a_new_user)",
                u"A_new_user (A_new_user)",
                u"b_new_user (b_new_user)",
                u"B_new_user (B_new_user)",
                u"new_user (new_user)",
                u"test_user_1_ (test_user_1_)",
            ],
        )
        self.assertEqual(
            [term.title for term in vocab2._terms],
            [u"anew User", u"A_new_user", u"b_new_user", u"B_new_user", u"new_user", u"test_user_1_"],
        )

    def test_YesNoForFacetedVocabulary(self):
        """ """
        vocab = get_vocab(self.portal, "imio.helpers.YesNoForFacetedVocabulary")
        self.assertEqual([term.value for term in vocab._terms], ["0", "1"])
        self.assertEqual([term.title for term in vocab._terms], ["yesno_value_false", "yesno_value_true"])


class EnhancedTermTests(SimpleTermTests):
    """Inherited class tests are also called"""

    def _getTargetClass(self):
        from imio.helpers.vocabularies import EnhancedTerm

        return EnhancedTerm

    def test_attrs(self):
        term = self._makeOne(u"value", title=u"title")
        self.assertDictEqual(term.attrs, {})
        term = self._makeOne(u"value", title=u"title", attr1=u"attr1")
        self.assertDictEqual(term.attrs, {"attr1": u"attr1"})



# -*- coding: utf-8 -*-

from imio.helpers import HAS_PLONE_6_AND_MORE
from imio.helpers import PLONE_MAJOR_VERSION
from imio.helpers.testing import FUNCTIONAL
from imio.helpers.testing import IntegrationTestCase
from imio.helpers.tests.utils import get_file_data
from imio.helpers.tests.utils import require_module
from imio.helpers.utils import create_image_content
from imio.helpers.xhtml import addClassToContent
from imio.helpers.xhtml import addClassToLastChildren
from imio.helpers.xhtml import imagesToData
from imio.helpers.xhtml import imagesToPath
from imio.helpers.xhtml import is_html
from imio.helpers.xhtml import markEmptyTags
from imio.helpers.xhtml import object_link
from imio.helpers.xhtml import removeBlanks
from imio.helpers.xhtml import removeCssClasses
from imio.helpers.xhtml import replace_content
from imio.helpers.xhtml import separate_images
from imio.helpers.xhtml import storeImagesLocally
from imio.helpers.xhtml import unescape_html
from imio.helpers.xhtml import xhtmlContentIsEmpty

import os
import six
import transaction


try:
    from Products.ATContentTypes.interfaces import IImageContent
except ImportError:
    print("require Archetypes")


picsum_image1_url = "https://fastly.picsum.photos/id/10/200/300.jpg?hmac=94QiqvBcKJMHpneU69KYg2pky8aZ6iBzKrAuhSUBB9s"
picsum_image2_url = "https://fastly.picsum.photos/id/1082/200/200.jpg?hmac=3usO1ziO7kCseIG52ruhRigxyk39W_L9eECWe1Hs6fY"

base64_img_data = (
    "data:image/png;base64,iVBORw0KGgoAAAANSUhEUgAAABQAAAAUCAYAAACNiR0NAAADWElEQVQ4ja2VX2yTBRTFj2QP"
    "JGtoLWWMZWN/tDBW1oQSoatuUNycNnWYMV1M9nVrJgRit6msVWuBj5fhQkhB5jagIqDQP2MmkKyDbToSKjSAo2yYInUhuiFEohghcQH7HR9M"
    "2RLWgJGTnJf7cJOb87v3AlOUk4OZGg1WlRqxztEE7/p6fL5hLbwf2dHneAfBTQ58k5YGLR5Hi56Fzirgs9JVaC41wpKdDV12NnQZGchXKpGp"
    "UiFDqcSsx2pWWIgS4wo4i4vw2pTyU//Bk1IqMcv0ElyaBVgGACrVMrNW6zy9dGlL9FHWaj/sTfQRRcwAAOifQ2XVajgBQC6fn1dWtu+ezXaM"
    "jY3J3bryXYqrW2gRPDEAeL1sbkW7Xd0BYAbW1+PT/HwYACAzc2XNzZt/SiQpSZO+e/ceJUkiSf7V18vfKwp5/9yLHBqKxgBgZ1NWx6gvj29X"
    "Zb4Fswmu+ekoAICsrCUlodA1qbY2QEHwUxD8bG7uYSh0jVu3DrCuLkCL4OflU3l0OWZy48aOW6mpqemly59+QVyb6wIAvLkGO9LTUfBMRoph"
    "sVpWaahuH9UXtbOg4gsWGHfdCgSG452dPSM225d3gsEIvd4wi5aX31arC6MajfXgQwkbV2CdWo2S1kb5xW8PzKNz9xoaWnZQ726jektXbHj4"
    "BjWa4vITJ777u7//Kq9c+ZVO587o87qFb5Qs0VQ8IGWBPA8AIJMhTfwAR7/aPnvstEfFru1zWNvpYGrbWabUu0+eDV+9bjJVWY8cCU6EQhFe"
    "uDDCzQ3O34LtjfRsssYUCoXik/dydw0fXHRDr5XrIJMhbcv76D7unv3z+UNzGPbmsHu/lrq2A0ypd58c+Pr762ZztbW7u3/i8OFBejx93Ptx"
    "a/KGiZH3i4qhvYdelsz73JLQtpkLd3cxVzwei0R+oU5XXD4+Ph4fHBzl2Ngf9Pl8yUeeGsq8xepKfXXnj0WGyVC83kvxPXuCI9u2Be5EIlGe"
    "OXORpldeTR7KdNjU1QVosfhpsfhpt/cwHP6JojjAmhovhRofL5/KTY7NdGBPhToBdjz+L9gTfb28/QDsHx4Ge5rVu2+zHWNDQ3InVk+YbvWe"
    "+HEAkp6v/6cnemATSvoCmtFvb0Kvy47BR72AfwDXsx4tZedcTQAAAABJRU5ErkJggg=="
)

base64_gif_img_data = (
    "data:image/gif;base64,R0lGODlhCgAKAPcAAP////79/f36+/3z8/zy8/rq7Prm6Pnq7Pje4vTx8v"
    "Pg5O6gqe2gqOq4v+igqt9tetxSYNs7Tdo5TNc5TNUbMdUbMNQRKNIKIdIJINIGHdEGHtEDGtACFdAAFtAAFNAAEs8AFAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA"
    "AAAAAAAAAAAAAAAAAAAAAAAAAAAAACIAIP3sAAoACBL2OAAAAPjHWRQAABUKiAAAABL2FAAAAhL+dPkBhPm4MP///xL2YPZNegA"
    "AAAAQAAAABgAAAAAAABL2wAAAAAAARRL25PEPfxQAAPEQAAAAAAAAA/3r+AAAAAAAGAAAABL2uAAAQgAAABL2pAAAAAAAAAAAAAA"
    "ADAAAAgABAfDTAAAAmAAAAAAAAAAAABL27Mku0AAQAAAAAAAAAEc1MUaDsAAAGBL3FPELyQAAAAABEgAAAwAAAQAAAxL22AAAmB"
    "L+dPO4dPPKQP///0c70EUoOQAAmMku0AAQABL3TAAAAEaDsEc1MUaDsAAABkT98AABEhL3wAAALAAAQAAAQBL3kQAACSH5BAEAA"
    "AAALAAAAAAKAAoAQAhGAAEIHEhw4IIIFCAsKCBwQAQQFyKCeKDAIEKFDAE4hBiRA0WCAwwUBLCAA4cMICg0YIjAQoaIFzJMSCCw"
    "5EkOKjM2FDkwIAA7"
)


class TestXHTMLModule(IntegrationTestCase):
    """
    Test all helper methods of xhtml module.
    """

    layer = FUNCTIONAL

    def test_removeBlanks(self):
        """
        Normal usecase of removeBlanks, this will remove
        every blank lines of a given XHTML content.
        """
        self.assertEqual(removeBlanks(""), "")
        self.assertEqual(removeBlanks(None), None)
        self.assertEqual(removeBlanks("<p></p>"), "")
        self.assertEqual(removeBlanks("<p></p><br><br /><br/>"), "")
        self.assertEqual(removeBlanks("<p>&nbsp;</p>"), "")
        self.assertEqual(removeBlanks("<p>&nbsp;<p><p>&nbsp;</p><i></i>"), "")
        self.assertEqual(removeBlanks("<p>Some text to keep</p><p>&nbsp;</p><i>&nbsp;</i>"), "<p>Some text to keep</p>")
        self.assertEqual(removeBlanks("<p> </p><p>Some text to keep</p><p>&nbsp;</p>"), "<p>Some text to keep</p>")
        self.assertEqual(removeBlanks("<p>Text line 1</p><p>Text line 2</p>"), "<p>Text line 1</p><p>Text line 2</p>")
        self.assertEqual(removeBlanks('<p><img src="my_image"/></p>'), '<p><img src="my_image"></p>')
        # complex tree filled
        self.assertEqual(
            removeBlanks("<ul><li>First line</li><li>second line</li></ul>"),
            "<ul><li>First line</li><li>second line</li></ul>",
        )
        # complex tree semi-filled
        self.assertEqual(
            removeBlanks("<ul><li>First line</li><li> </li></ul>"), "<ul><li>First line</li><li> </li></ul>"
        )
        # empty complex tree, is not wiped out because master tag containing children
        self.assertEqual(
            removeBlanks("<ul><li>&nbsp;</li><li>&nbsp;</li></ul>"), "<ul><li>&#160;</li><li>&#160;</li></ul>"
        )
        # result is not prettyfied (pretty_print=False) but if source was
        # pretty, then the result is pretty also
        self.assertEqual(
            removeBlanks("<ul>\n  <li>&nbsp;</li>\n  <li>&nbsp;</li>\n</ul>\n"),
            "<ul>\n  <li>&#160;</li>\n  <li>&#160;</li>\n</ul>\n",
        )

    def test_xhtmlContentIsEmpty(self):
        """
        Test if a given XHTML content will produce someting on render.
        """
        self.assertTrue(xhtmlContentIsEmpty(""))
        self.assertTrue(xhtmlContentIsEmpty(None))
        self.assertTrue(xhtmlContentIsEmpty("<p></p>"))
        self.assertTrue(xhtmlContentIsEmpty("<br/><br/><br/>"))
        self.assertTrue(xhtmlContentIsEmpty("<p></p><p>&nbsp;</p><p> </p><b></b><i> </i>"))
        self.assertTrue(xhtmlContentIsEmpty("<img />"))
        # empty <p> with an attribute, considered not empty
        self.assertTrue(not xhtmlContentIsEmpty('<p class="special_empty_line"></p>'))
        # tree with children, considered not empty
        self.assertTrue(not xhtmlContentIsEmpty("<table><tr><td> </td><td>&nbsp;</td></tr></table>"))
        self.assertTrue(not xhtmlContentIsEmpty("<p><ul><li> </li><li>&nbsp;</li></ul></p>"))
        # tree with some kind of children are considered empty
        self.assertTrue(xhtmlContentIsEmpty("<p><br /></p>"))
        self.assertTrue(xhtmlContentIsEmpty("<p><br /><br /></p>"))
        self.assertTrue(xhtmlContentIsEmpty("<p>&nbsp;<br /></p>"))
        self.assertTrue(xhtmlContentIsEmpty("<p></p>\n\n"))
        self.assertTrue(xhtmlContentIsEmpty("<p><br />\n&nbsp;</p>\n"))

        # other content
        self.assertTrue(not xhtmlContentIsEmpty("Some text without any tag"))
        self.assertTrue(not xhtmlContentIsEmpty("<p>Some broken HTML<broken_tag>text<broken_tag> </i></div>"))
        self.assertTrue(not xhtmlContentIsEmpty("<p>Some text to keep</p>"))
        self.assertTrue(not xhtmlContentIsEmpty("<p>&nbsp;</p><p>Some text to keep</p><i>&nbsp;</i>"))
        self.assertTrue(not xhtmlContentIsEmpty("<p>&nbsp;</p><i>Some text to keep</i>"))
        self.assertTrue(not xhtmlContentIsEmpty("<table><tr><td>Some text to keep</td><td>&nbsp;</td></tr></table>"))
        self.assertTrue(not xhtmlContentIsEmpty('<p><img src="my_image_path.png"></p>'))

    def test_addClassToContent(self):
        """
        This will add a CSS class to every CONTENT_TAGS find in the given xhtmlContent.
        """
        self.assertEqual(addClassToContent("<p>My text</p>", css_class="sample"), '<p class="sample">My text</p>')
        self.assertEqual(addClassToContent("", css_class="sample"), "")
        self.assertEqual(addClassToContent(None, css_class="sample"), None)
        self.assertEqual(addClassToContent("text", css_class="sample"), "text")
        # if tag is not handled, it does not change anything, by default, tags 'p' and 'li' are handled
        self.assertEqual(
            addClassToContent("<unhandled_tag>My text with tag not handled</unhandled_tag>", css_class="sample"),
            "<unhandled_tag>My text with tag not handled</unhandled_tag>",
        )
        # handled tag
        self.assertEqual(addClassToContent("<p>My text</p>", css_class="sample"), '<p class="sample">My text</p>')
        # already existing class
        self.assertEqual(
            addClassToContent('<p class="another">My text</p>', css_class="sample"),
            '<p class="sample another">My text</p>',
        )
        self.assertEqual(
            addClassToContent(
                "<p>My text</p>"
                '<p class="class1 class2">My text</p>'
                "<p>My text</p>"
                '<p class="class3">My text</p>'
                "<unhandled_tag>My text</unhandled_tag>"
                "<span>My text</span>"
                "<p>My text</p>",
                css_class="sample",
            ),
            '<p class="sample">My text</p>'
            '<p class="sample class1 class2">My text</p>'
            '<p class="sample">My text</p>'
            '<p class="sample class3">My text</p>'
            "<unhandled_tag>My text</unhandled_tag>"
            '<span class="sample">My text</span>'
            '<p class="sample">My text</p>',
        )

    def test_addClassToLastChildren(self):
        """
        Test if adding a class to last x tags of a given XHTML content works.
        By default, this method receives following parameters :
        - xhtmlContent;
        - classNames={'p': 'ParaKWN',
                     'li': 'podItemKeepWithNext'};
        - tags=('p', 'li', );
        - numberOfChars=180.
        """
        self.assertEqual(addClassToLastChildren(""), "")
        self.assertEqual(addClassToLastChildren(None), None)
        self.assertEqual(addClassToLastChildren("text"), "text")
        # if tag is not handled, it does not change anything, by default, tags 'p' and 'li' are handled
        self.assertEqual(
            addClassToLastChildren("<unhandled_tag>My text with tag not handled</unhandled_tag>"),
            "<unhandled_tag>My text with tag not handled</unhandled_tag>",
        )
        # now test with a single small handled tag, text size is lower than numberOfChars
        self.assertEqual(addClassToLastChildren("<p>My small text</p>"), '<p class="ParaKWN">My small text</p>')
        # existing class attribute is kept
        self.assertEqual(
            addClassToLastChildren('<p class="myclass">My small text</p>'),
            '<p class="ParaKWN myclass">My small text</p>',
        )
        # test that if text is smaller than numberOfChars, several last tags are adapted
        self.assertEqual(
            addClassToLastChildren("<p>My small text</p><p>My small text</p>"),
            '<p class="ParaKWN">My small text</p>' '<p class="ParaKWN">My small text</p>',
        )
        # large text, only relevant tags are adapted until numberOfChars is rechead
        self.assertEqual(
            addClassToLastChildren(
                "<p>13 chars line</p><p>33 characters text line text line</p><p>33 characters text line text line</p>",
                numberOfChars=60,
            ),
            '<p>13 chars line</p><p class="ParaKWN">33 characters text line text line</p>'
            '<p class="ParaKWN">33 characters text line text line</p>',
        )
        # tags with children
        self.assertEqual(
            addClassToLastChildren(
                "<p>First untouched paragraph</p>"
                "<p><strong>Strong text Strong text Strong text Strong text</strong> "
                "Paragraph text Paragraph text Paragraph text</p>",
                numberOfChars=60,
            ),
            "<p>First untouched paragraph</p>"
            '<p class="ParaKWN"><strong>Strong text Strong text Strong text Strong text</strong> '
            "Paragraph text Paragraph text Paragraph text</p>",
        )
        # test mixing different handled tags like 'ul', 'li' and 'p'
        self.assertEqual(
            addClassToLastChildren(
                "<p>13 chars line</p><ul><li>Line 1</li><li>Line 2</li><li>33 characters text line text line</li></ul>"
                "<p>33 characters text line text line</p>",
                numberOfChars=60,
            ),
            '<p>13 chars line</p><ul class="podBulletItemKeepWithNext">'
            '<li>Line 1</li><li>Line 2</li><li class="podItemKeepWithNext">33 '
            'characters text line text line</li></ul><p class="ParaKWN">33 '
            "characters text line text line</p>",
        )
        # style is also applied on <ol>
        self.assertEqual(
            addClassToLastChildren("<ol><li>Line 1</li><li>Line 2</li></ol>"),
            '<ol><li class="podItemKeepWithNext">Line 1</li><li class="podItemKeepWithNext">Line 2</li></ol>',
        )
        # as soon as an unhandled tag is discover, adaptation stops
        self.assertEqual(
            addClassToLastChildren('<p>13 chars line</p><img src="image.png"><p>13 chars line</p>'),
            '<p>13 chars line</p><img src="image.png"><p class="ParaKWN">13 chars line</p>',
        )
        # test with sub tags
        self.assertEqual(
            addClassToLastChildren("<p>Text</p><p><u><strong>dsdklm</strong></u></p>"),
            '<p class="ParaKWN">Text</p><p class="ParaKWN"><u><strong>dsdklm</strong></u></p>',
        )
        # test every available tags
        self.assertEqual(
            addClassToLastChildren(
                "<p>Text</p>"
                "<p><strong>Strong</strong></p>"
                "<p><span>Span</span></strong>"
                "<p><strike>Strike</strike></p>"
                "<p><i>I</i></p>"
                "<p><em>Em</em></p>"
                "<p><u>U</u></p>"
                "<p><small>Small</small></p>"
                "<p><mark>Mark</mark></p>"
                "<p><del>Del</del></p>"
                "<p><ins>Ins</ins></p>"
                "<p><sub>Sub</sub></p>"
                "<p><sup>Sup</sup></p>",
                numberOfChars=85,
            ),
            '<p class="ParaKWN">Text</p>'
            '<p class="ParaKWN"><strong>Strong</strong></p>'
            '<p class="ParaKWN"><span>Span</span></p>'
            '<p class="ParaKWN"><strike>Strike</strike></p>'
            '<p class="ParaKWN"><i>I</i></p>'
            '<p class="ParaKWN"><em>Em</em></p>'
            '<p class="ParaKWN"><u>U</u></p>'
            '<p class="ParaKWN"><small>Small</small></p>'
            '<p class="ParaKWN"><mark>Mark</mark></p>'
            '<p class="ParaKWN"><del>Del</del></p>'
            '<p class="ParaKWN"><ins>Ins</ins></p>'
            '<p class="ParaKWN"><sub>Sub</sub></p>'
            '<p class="ParaKWN"><sup>Sup</sup></p>',
        )
        # does not break with unknown sub tags
        self.assertEqual(
            addClassToLastChildren("<p><unknown><tagtag>Text</tagtag></unknown></p>"),
            '<p class="ParaKWN"><unknown><tagtag>Text</tagtag></unknown></p>',
        )
        # special characters are transformed to HTML entities
        text = "<p>Some spécial charaçters &eacute;</p><p>&nbsp;</p>"
        self.assertEqual(
            addClassToLastChildren(text),
            '<p class="ParaKWN">Some sp&#233;cial chara&#231;ters &#233;</p>' '<p class="ParaKWN">&#160;</p>',
        )
        # result is not prettyfied (pretty_print=False) but if source was
        # pretty, then the result is pretty also
        self.assertEqual(
            addClassToLastChildren('<p>text</p>\n<img src="image.png"/>\n<p>text</p>\n'),
            '<p>text</p>\n<img src="image.png">\n<p class="ParaKWN">text</p>\n',
        )

    def test_markEmptyTags(self):
        """
        Test if empty tags are correctly marked with a CSS class.
        markEmptyTags can receive different parameters :
        - markingClass : the class that will mark the empty tags;
        - tagTitle : if given, a 'title' attribute is added to the marked tags;
        - onlyAtTheEnd : if True, only empty tags at the end of the XHTML content are marked;
        - tags : a list of handled tags.
        """
        self.assertEqual(markEmptyTags(""), "")
        self.assertEqual(markEmptyTags(None), None)
        self.assertEqual(markEmptyTags("<p></p>"), '<p class="highlightBlankRow"></p>')
        self.assertEqual(markEmptyTags("<p> </p>"), '<p class="highlightBlankRow"> </p>')
        if six.PY3:
            self.assertEqual(markEmptyTags("<p>&nbsp;</p>"), '<p class="highlightBlankRow">\xa0</p>')
            self.assertEqual(markEmptyTags("<span>&nbsp;</span>"), "<span>\xa0</span>")
            # result is not prettyfied (pretty_print=False) but if source was
            # pretty, then the result is pretty also
            self.assertEqual(
                markEmptyTags("<p>String with special chars: é</p>\n"), "<p>String with special chars: é</p>\n"
            )
        else:
            self.assertEqual(markEmptyTags("<p>&nbsp;</p>"), '<p class="highlightBlankRow">\xc2\xa0</p>')
            self.assertEqual(markEmptyTags("<span>&nbsp;</span>"), "<span>\xc2\xa0</span>")
            # result is not prettyfied (pretty_print=False) but if source was
            # pretty, then the result is pretty also
            self.assertEqual(
                markEmptyTags(u"<p>Unicode string with special chars: \xe9</p>\n"),
                "<p>Unicode string with special chars: \xc3\xa9</p>\n",
            )

        self.assertEqual(markEmptyTags("<p>Text</p>"), "<p>Text</p>")
        self.assertEqual(markEmptyTags("<p>Text</p><p></p>"), '<p>Text</p><p class="highlightBlankRow"></p>')
        # change markingClass
        self.assertEqual(markEmptyTags("<p> </p>", markingClass="customClass"), '<p class="customClass"> </p>')
        # "span" not handled by default
        self.assertEqual(markEmptyTags("<span></span>"), "<span></span>")
        # but "span" could be handled if necessary
        self.assertEqual(markEmptyTags("<span></span>", tags=("span",)), '<span class="highlightBlankRow"></span>')
        # by default every empty tags are highlighted, but we can specify to highlight only trailing
        # ones aka only tags ending the XHTML content
        self.assertEqual(
            markEmptyTags("<p></p><p>Text</p><p></p>"),
            '<p class="highlightBlankRow"></p><p>Text</p><p class="highlightBlankRow"></p>',
        )
        self.assertEqual(
            markEmptyTags("<p></p><p>Text</p><p></p>", onlyAtTheEnd=True),
            '<p></p><p>Text</p><p class="highlightBlankRow"></p>',
        )
        # we can also specify to add a title attribute to the highlighted tags
        self.assertEqual(
            markEmptyTags("<p>Text</p><p></p>", tagTitle="My tag title"),
            '<p>Text</p><p class="highlightBlankRow" title="My tag title"></p>',
        )
        # if an empty tag already have a class, the marking class is appended to it
        self.assertEqual(
            markEmptyTags("<p class='existingClass'></p>"), '<p class="highlightBlankRow existingClass"></p>'
        )

        # we may mark Unicode as well as UTF-8 xhtmlContent
        self.assertEqual(
            markEmptyTags("<p>UTF-8 string with special chars: \xc3\xa9</p>"),
            "<p>UTF-8 string with special chars: \xc3\xa9</p>",
        )

    def test_removeCssClasses(self):
        """
        Test if given CSS classes are removed from entire xhtmlContent.
        """
        self.assertEqual(removeCssClasses(""), "")
        self.assertEqual(removeCssClasses("", css_classes=["my_class"]), "")
        self.assertEqual(removeCssClasses(None), None)
        self.assertEqual(removeCssClasses(None, css_classes=["my_class"]), None)
        self.assertEqual(removeCssClasses("<p></p>", css_classes=["my_class"]), "<p></p>")
        self.assertEqual(removeCssClasses("<p>Text</p>", css_classes=["my_class"]), "<p>Text</p>")
        self.assertEqual(
            removeCssClasses('<p class="another_class">Text</p>', css_classes=["my_class"]),
            '<p class="another_class">Text</p>',
        )
        self.assertEqual(removeCssClasses('<p class="my_class">Text</p>', css_classes=["my_class"]), "<p>Text</p>")
        self.assertEqual(
            removeCssClasses('<p><span class="my_class">Text</span></p>', css_classes=["my_class"]),
            "<p><span>Text</span></p>",
        )
        self.assertEqual(
            removeCssClasses(
                "<p>"
                '<span class="my_class">Text</span>'
                '<span class="another_class">Text</span>'
                '<span class="my_class">Text</span>'
                "</p>",
                css_classes=["my_class"],
            ),
            "<p>" "<span>Text</span>" '<span class="another_class">Text</span>' "<span>Text</span>" "</p>",
        )
        self.assertEqual(
            removeCssClasses(
                "<ul>" '<li class="another_class">Text</li>' '<li class="my_class">Text</li>' "</ul>",
                css_classes=["my_class"],
            ),
            "<ul>" '<li class="another_class">Text</li>' "<li>Text</li>" "</ul>",
        )
        self.assertEqual(
            removeCssClasses(
                "<ul>"
                '<li class="another_class"><span class="my_class">Text</span></li>'
                '<li class="my_class">Text</li>'
                "</ul>",
                css_classes=["my_class"],
            ),
            "<ul>" '<li class="another_class"><span>Text</span></li>' "<li>Text</li>" "</ul>",
        )

    def test_imagesToPath(self):
        """
        Test that images src contained in a XHTML content are correctly changed to
        the blob file system absolute path.
        """
        # create a document and an image
        docId = self.portal.invokeFactory("Document", id="doc", title="Document")
        doc = getattr(self.portal, docId)
        img_without_blob_id = self.portal.invokeFactory("Image", id="img_without_blob", title="Image")
        img_without_blob = getattr(self.portal, img_without_blob_id)
        # no blob
        self.assertEqual(img_without_blob.get_size(), 0)
        data = get_file_data("dot.gif")
        img = create_image_content(self.portal, "Image", "img", data, filename="dot.gif")
        transaction.savepoint()
        # has a blob
        self.assertEqual(img.get_size(), 873)
        if HAS_PLONE_6_AND_MORE:
            blob = img.image._blob
        else:
            blob = img.getBlobWrapper().blob
        blob._p_activate()
        img_blob_path = blob._p_blob_committed
        # folder with doc2 to test relative path
        subfolderId = self.portal.invokeFactory("Folder", id="subfolder", title="Folder")
        subfolder = getattr(self.portal, subfolderId)
        doc2Id = subfolder.invokeFactory("Document", id="doc2", title="Document")
        doc2 = getattr(subfolder, doc2Id)

        # we do not setText because the mutator is overrided to use mxTidy
        # to prettify the HTML
        # test with internal image, absolute path
        text = '<p>Image absolute path <img src="{0}/img"> end of text.</p>'.format(self.portal_url)
        expected = text.replace("{0}/img".format(self.portal_url), img_blob_path)
        self.assertEqual(imagesToPath(doc, text).strip(), expected)
        # if we use an image having no blob, nothing is changed
        text = '<p>Image absolute path <img src="{0}/img_without_blob"> end of text.</p>'.format(self.portal_url)
        self.assertEqual(imagesToPath(doc, text).strip(), text)

        # test with internal image, relative path
        text = '<p>Image relative path <img src="../img" alt="Image" title="Image"> end of text.</p>'
        expected = text.replace("../img", img_blob_path)
        self.assertEqual(imagesToPath(doc2, text).strip(), expected)
        # if we use an image having no blob, nothing is changed
        text = '<p>Image relative path <img src="../img_without_blob" alt="Image" title="Image"> end of text.</p>'
        self.assertEqual(imagesToPath(doc2, text).strip(), text)

        # test with src to an ImageScale instead the full image
        text = (
            '<p>Link to ImageScale absolute path <img src="{0}/img/image_preview"> '
            'and relative path <img src="../img/image_preview">.</p>'.format(self.portal_url)
        )
        expected = text.replace("{0}/img/image_preview".format(self.portal_url), img_blob_path)
        expected = expected.replace("../img/image_preview", img_blob_path)
        if PLONE_MAJOR_VERSION < 5:
            self.assertEqual(imagesToPath(doc2, text).strip(), expected)

        # test with src to image that is not an image, src will be to doc
        text = (
            '<p>Src image to doc absolute path <img src="{0}/doc"> '
            'and relative path <img src="../doc">.</p>'.format(self.portal_url)
        )
        # left as is
        self.assertEqual(imagesToPath(doc2, text).strip(), text)

        # external image, absolute_path
        # nothing changed
        text = '<p>Image absolute path <img src="http://www.othersite.com/image.png"> end of text.</p>'
        self.assertEqual(imagesToPath(doc, text).strip(), text)

        # image that does not exist anymore, absolute and relative path
        text = (
            '<p>Removed image absolute path <img src="{0}/removed_img"> '
            'and relative path <img src="../removed_img">.</p>'.format(self.portal_url)
        )
        # left as is
        self.assertEqual(imagesToPath(doc, text).strip(), text)

        # more complex case with html sublevels, relative and absolute path images
        text = """<p>Image absolute path <img src="{0}/img"> end of text.</p>
<div>
  <p>Image absolute path2 same img <img src="{0}/img"> end of text.</p>
</div>
<p>Image absolute path <img src="http://www.othersite.com/image.png"> end of text.</p>
<div>
  <p>
    <strong>Image relative path <img src="../img" alt="Image" title="Image"> end of text.</strong>
  </p>
</div>
<p>Removed image absolute path <img src="{0}/removed_img" alt="Image" title="Image"> end of text.</p>
<p>Removed image relative path <img src="../removed_img" alt="Image" title="Image"> end of text.</p>""".format(
            self.portal_url
        )
        expected = text.replace("{0}/img".format(self.portal_url), img_blob_path)
        expected = expected.replace("../img", img_blob_path)
        self.assertEqual(imagesToPath(doc2, text).replace("\n", ""), expected.replace("\n", ""))

        # does not break if xhtmlContent is empty
        self.assertEqual(imagesToPath(doc, ""), "")

        # if text does not contain images, it is returned as is
        text = "<p>Text without images.</p>"
        self.assertEqual(imagesToPath(doc, text), text)

        # image outside any other tag
        text = '<img src="{0}/img/image_preview"><img src="../img/image_preview">'.format(self.portal_url)
        expected = text.replace("{0}/img/image_preview".format(self.portal_url), img_blob_path)
        expected = expected.replace("../img/image_preview", img_blob_path)
        if PLONE_MAJOR_VERSION < 5:
            self.assertEqual(imagesToPath(doc2, text).replace("\n", ""), expected.replace("\n", ""))

        # image without src, nothing done
        text = '<img title="My image without src">'
        self.assertEqual(imagesToPath(doc, text), text)

        # using resolveuid and absolute path and relative path
        text = (
            '<img src="resolveuid/{0}/image_preview" alt="Image" title="Image">'
            '<img src="resolveuid/{0}" alt="Image" title="Image">'
            '<img src="{1}/img/image_preview">'
            '<img src="../img/image_preview">'.format(img.UID(), self.portal_url)
        )
        expected = text.replace("resolveuid/{0}/image_preview".format(img.UID()), img_blob_path)
        expected = expected.replace("resolveuid/{0}".format(img.UID()), img_blob_path)
        expected = expected.replace("{0}/img/image_preview".format(self.portal_url), img_blob_path)
        expected = expected.replace("../img/image_preview", img_blob_path)
        if PLONE_MAJOR_VERSION < 5:
            self.assertEqual(imagesToPath(doc2, text).replace("\n", ""), expected.replace("\n", ""))

        # does not break with wrong resolveuid
        if six.PY2:
            text = '<img src="resolveuid/unknown_uid" alt="Image" title="Image">'
            self.assertEqual(imagesToPath(doc2, text), text)
        else:
            text = '<img alt="Image" src="resolveuid/unknown_uid" title="Image">'
            self.assertEqual(imagesToPath(doc2, text), text)

    def test_imagesToData(self):
        """
        Test that images src contained in a XHTML content are correctly changed to
        the a data base64 value.
        Method is based on same as imagesToPath so we do not redo
        tests that are already done in test_imagesToPath.
        """
        # create a document and an image
        docId = self.portal.invokeFactory("Document", id="doc", title="Document")
        doc = getattr(self.portal, docId)
        data = get_file_data("dot.gif")
        img = create_image_content(self.portal, "Image", "img", data, filename="dot.gif")
        # has a blob
        self.assertEqual(img.get_size(), 873)
        text = '<p>Image <img src="{0}/img"> end of text.</p>'.format(self.portal_url)
        transaction.savepoint()
        self.assertEqual(
            imagesToData(doc, text), '<p>Image <img src="{0}"> end of text.</p>'.format(base64_gif_img_data)
        )

    @require_module("Products.Archetypes")
    def test_storeExternalImagesLocally(self):
        """
        Test that images src contained in a XHTML that reference external images is
        updated correctly and external image is stored locally.
        """
        # create a document and an image
        docId = self.portal.invokeFactory("Document", id="doc", title="Document")
        doc = getattr(self.portal, docId)
        data = get_file_data("dot.gif")
        img = create_image_content(self.portal, "Image", "img", data, filename="dot.gif")

        # does not break if xhtmlContent is empty
        self.assertEqual(storeImagesLocally(doc, ""), "")
        self.assertEqual(storeImagesLocally(doc, "<p>&nbsp;</p>"), "<p>&nbsp;</p>")
        # not changed if only contains local images, using relative or absolute path
        text = '<p>Image absolute path <img src="/img"/> and relative path <img src="../img"/>.</p>'
        self.assertEqual(storeImagesLocally(doc, text), text)

        # link to unexisting external image, site exists but not image (error 404) nothing changed
        text = '<p>Unexisting external image <img src="http://www.imio.be/unexistingimage.png"/></p>.'
        self.assertEqual(storeImagesLocally(doc, text, replace_not_found_image=False), text)

        # link to unexisting site, nothing changed
        text = '<p>Unexisting external site <img src="http://www.unexistingsite.be/unexistingimage.png"/>.</p>'
        self.assertEqual(storeImagesLocally(doc, text, replace_not_found_image=False), text)

        # when using replace_not_found_image=True image is replaced by a "Not found" image
        text = '<p>Unexisting external image <img src="http://nohost/plone/imagenotfound.jpg"></p>.'
        self.assertEqual(storeImagesLocally(doc, text), text)
        # calling it a second time will store it again because doc is not a container
        text = '<p>Unexisting external image <img src="http://nohost/plone/imagenotfound-1.jpg"></p>.'
        self.assertEqual(storeImagesLocally(doc, text), text)

        # working example
        text = '<p>Working external image <img src="%s"/>.</p>' % picsum_image1_url
        # we have Content-Dispsition header
        downloaded_img_path, downloaded_img_infos = six.moves.urllib.request.urlretrieve(picsum_image1_url)
        self.assertTrue(downloaded_img_infos.getheader("Content-Disposition"))
        # image object does not exist for now
        self.assertFalse("10-200x300.jpg" in self.portal.objectIds())
        self.assertEqual(
            storeImagesLocally(doc, text),
            '<p>Working external image <img src="{0}/10-200x300.jpg">.</p>'.format(self.portal_url),
        )
        img = self.portal.get("10-200x300.jpg")
        self.assertTrue(IImageContent.providedBy(img))

        # working example with a Folder, this test case where we have a container
        # using RichText field, in this case the Image is stored in the Folder, not next to it
        text = '<p>Working external image <img src="%s">.</p>' % picsum_image2_url
        expected = '<p>Working external image <img src="{0}/folder/1082-200x200.jpg">.</p>'.format(self.portal_url)
        self.assertFalse("1082-200x200.jpg" in self.portal.folder.objectIds())
        self.assertEqual(storeImagesLocally(self.portal.folder, text), expected)
        mascotte = self.portal.folder.get("1082-200x200.jpg")
        self.assertTrue(IImageContent.providedBy(mascotte))

        # link to external image without Content-Disposition
        # it is downloaded nevertheless but used filename will be 'image-1'
        text = '<p>External site <img src="http://www.imio.be/logo.png">.</p>'
        expected = '<p>External site <img src="{0}/folder/image-1.png">.</p>'.format(self.portal_url)
        downloaded_img_path, downloaded_img_infos = six.moves.urllib.request.urlretrieve("http://www.imio.be/logo.png")
        self.assertIsNone(downloaded_img_infos.getheader("Content-Disposition"))
        self.assertEqual(storeImagesLocally(self.portal.folder, text), expected)
        logo = self.portal.folder.get("image-1.png")
        self.assertTrue(IImageContent.providedBy(logo))

        # if context is a container, an "Not found" image is only added one time
        testFolderId = self.portal.invokeFactory("Folder", id="testfolder", title="Test folder")
        testfolder = getattr(self.portal, testFolderId)
        text = '<p>Unexisting external site <img src="http://www.unexistingsite.be/unexistingimage.png"/>.</p>'
        expected = '<p>Unexisting external site <img src="http://nohost/plone/testfolder/imagenotfound.jpg">.</p>'
        self.assertEqual(testfolder.objectIds(), [])
        self.assertEqual(storeImagesLocally(testfolder, text), expected)
        self.assertEqual(testfolder.objectIds(), ["imagenotfound.jpg"])
        # calling it a second time with a not found image will retrieve it again
        expected = '<p>Unexisting external site <img src="http://nohost/plone/testfolder/imagenotfound-1.jpg">.</p>'
        self.assertEqual(storeImagesLocally(testfolder, text), expected)
        self.assertEqual(testfolder.objectIds(), ["imagenotfound.jpg", "imagenotfound-1.jpg"])
        # but if image found, nothing is done
        text = '<p>Unexisting external site <img src="http://nohost/plone/testfolder/imagenotfound.jpg">.</p>'
        self.assertEqual(storeImagesLocally(testfolder, text), text)
        self.assertEqual(testfolder.objectIds(), ["imagenotfound.jpg", "imagenotfound-1.jpg"])

    def test_storeExternalImagesLocallyWithResolveUID(self):
        """ """
        # working example
        text = '<p>Working external image <img src="%s">.</p>' % picsum_image1_url
        result = storeImagesLocally(self.portal, text, force_resolve_uid=True)

        # image was downloaded and link to it was turned to a resolveuid
        img = self.portal.get("10-200x300.jpg")
        self.assertEqual(result, '<p>Working external image <img src="resolveuid/{0}">.</p>'.format(img.UID()))

    @require_module("Products.Archetypes")
    def test_storeInternalImagesLocally(self):
        """
        Test that images src contained in a XHTML that reference internal images stored
        in another context are stored in current context.
        """
        data = get_file_data("dot.gif")
        create_image_content(self.portal, "Image", "dot.gif", data, filename="dot.gif")
        text = '<p>Internal image <img src="{0}/dot.gif">.</p>'.format(self.portal_url)
        expected = '<p>Internal image <img src="{0}/folder/dot.gif">.</p>'.format(self.portal_url)
        # image was created in folder
        self.assertEqual(storeImagesLocally(self.portal.folder, text), expected)
        image = self.portal.folder.get("dot.gif")
        self.assertTrue(IImageContent.providedBy(image))

        # nothing changed if image already stored to context
        self.assertEqual(storeImagesLocally(self.portal.folder, expected), expected)

        # now check when internal image does not exist
        text = '<p>Internal image <img src="{0}/unknown.gif">.</p>'.format(self.portal_url)
        # with replace_not_found_image=False in case an internal image is not found, nothing is done
        self.assertEqual(storeImagesLocally(self.portal.folder, text, replace_not_found_image=False), text)
        text = '<p>Internal image <img src="http://nohost/plone/folder/imagenotfound.jpg">.</p>'
        self.assertEqual(storeImagesLocally(self.portal.folder, text, replace_not_found_image=True), text)

    def test_storeInternalImagesLocallyWithResolveUID(self):
        """ """
        file_path = os.path.join(os.path.dirname(__file__), "dot.gif")
        fd = open(file_path, "rb")
        data = fd.read()
        create_image_content(self.portal, "Image", "dot.gif", data=data)
        img2 = create_image_content(self.portal, "Image 2", "dot2.gif", data=data)
        img3 = create_image_content(self.portal, "Image 3", "dot3.gif", data=data)
        img4 = create_image_content(self.portal, "Image 4", "dot4.gif", data=data)
        img5 = create_image_content(self.portal, "Image 5", "dot5.gif", data=data)
        fd.close()
        text = (
            '<p>Int img full url <img src="{0}/dot.gif"/>.</p>'
            '<p>Int img resolveuid <img src="resolveuid/{1}"/>.</p>'
            '<p>Int img resolveuid and image_preview <img src="resolveuid/{2}/image_preview"/>.</p>'
            '<p>Int img resolveuid portal_url <img src="{0}/resolveuid/{3}"/>.</p>'
            '<p>Int img resolveuid portal_url image_preview <img src="{0}/resolveuid/{4}/image_preview"/>.</p>'.format(
                self.portal_url, img2.UID(), img3.UID(), img4.UID(), img5.UID()
            )
        )

        # every images uri are turned to resolveuid
        result = storeImagesLocally(self.portal.folder, text, force_resolve_uid=True)
        # images were created in folder
        self.assertEqual(self.portal.folder.objectIds(), ["dot.gif", "dot2.gif", "dot3.gif", "dot4.gif", "dot5.gif"])
        new_img = self.portal.folder.get("dot.gif")
        new_img2 = self.portal.folder.get("dot2.gif")
        new_img3 = self.portal.folder.get("dot3.gif")
        new_img4 = self.portal.folder.get("dot4.gif")
        new_img5 = self.portal.folder.get("dot5.gif")
        expected = (
            '<p>Int img full url <img src="resolveuid/{0}">.</p>'
            '<p>Int img resolveuid <img src="resolveuid/{1}">.</p>'
            '<p>Int img resolveuid and image_preview <img src="resolveuid/{2}/image_preview">.</p>'
            '<p>Int img resolveuid portal_url <img src="resolveuid/{3}">.</p>'
            '<p>Int img resolveuid portal_url image_preview <img src="resolveuid/{4}/image_preview">.</p>'.format(
                new_img.UID(), new_img2.UID(), new_img3.UID(), new_img4.UID(), new_img5.UID()
            )
        )
        self.assertEqual(result, expected)

    def test_storeImagesLocallyWithWrongSrc(self):
        """Do not fail when no src in the <img> tag or src is to something else than an image."""
        # no src
        text = '<p>Text <img style="width: 50px; height: 50px;"/>.</p>'
        # nothing was changed
        self.assertEqual(storeImagesLocally(self.portal.folder, text), text)

        # src to a non Image, Folder
        text = '<p>Text <img src="{0}"style="width: 50px; height: 50px;"/>.</p>'.format(self.portal.absolute_url())
        # nothing was changed
        self.assertEqual(storeImagesLocally(self.portal.folder, text, replace_not_found_image=False), text)

        # src to a non Image, Portal
        text = '<p>Text <img src="{0}"style="width: 50px; height: 50px;"/>.</p>'.format(
            self.portal.folder.absolute_url()
        )
        # nothing was changed
        self.assertEqual(storeImagesLocally(self.portal.folder2, text, replace_not_found_image=False), text)

    def test_storeImagesLocallyWithImgPathStoredInFolderStartingWithContextURL(self):
        """Make sure an image stored in a folder that have same id as begining
        of another folder the image is stored in downloads the image."""

        self.assertTrue(self.portal.folder2.absolute_url().startswith(self.portal.folder.absolute_url()))

        text = '<p>Working external image <img src="%s">.</p>' % picsum_image1_url
        storeImagesLocally(self.portal.folder2, text)
        text = '<p>Working external image <img src="{0}">.</p>'.format(
            self.portal.folder2.objectValues()[0].absolute_url()
        )
        self.assertFalse(self.portal.folder.objectIds())
        storeImagesLocally(self.portal.folder, text)
        self.assertTrue(self.portal.folder.objectIds())

    def test_storeExternalImagesLocallyImageURLWithNonASCIIChars(self):
        """URL to image may contain non ASCII chars."""
        text = '<p>External image with special chars <img src="http://www.imio.be/émage.png">.</p>'
        result = storeImagesLocally(self.portal, text, force_resolve_uid=True, replace_not_found_image=False)
        # in this case, image is ignored
        self.assertEqual(result, text)

    def test_storeExternalImagesLocallyDataBase64Image(self):
        """Test when src is a data:image base64 encoded image."""
        if six.PY3:
            self.skipTest(
                "Skipped for py3: storeImagesLocally broken with base64 image, downloaded_img_infos is "
                "<email.message.Message object"
            )
        text = (
            "<p>Image using data:image base64 encoded binary.</p>"
            '<p><img src="{0}" /></p><p><img src="{0}" /></p>'.format(base64_img_data)
        )
        expected = (
            "<p>Image using data:image base64 encoded binary.</p>"
            '<p><img src="http://nohost/plone/image-1.png"></p>'
            '<p><img src="http://nohost/plone/image-2.png"></p>'
        )
        # for now, no images
        self.assertEqual(len([obj for obj in self.portal.objectValues() if obj.portal_type == "Image"]), 0)
        result = storeImagesLocally(self.portal, text)
        self.assertEqual(result, expected)
        self.assertEqual(len([obj for obj in self.portal.objectValues() if obj.portal_type == "Image"]), 2)

    def test_object_link(self):
        obj = self.portal.folder
        obj.setTitle(u"Folder title")
        self.assertEqual(object_link(obj), u'<a href="http://nohost/plone/folder/view">Folder title</a>')
        self.assertEqual(object_link(obj, view="edit"), u'<a href="http://nohost/plone/folder/edit">Folder title</a>')
        self.assertEqual(
            object_link(obj, attribute="portal_type"), u'<a href="http://nohost/plone/folder/view">Folder</a>'
        )
        self.assertEqual(
            object_link(obj, view="edit", attribute="Description"), u'<a href="http://nohost/plone/folder/edit"></a>'
        )
        self.assertEqual(
            object_link(obj, view="edit", attribute="Description", target="_blank"),
            u'<a href="http://nohost/plone/folder/edit" target="_blank"></a>',
        )
        obj.setTitle(u"Folder title <script />")
        self.assertEqual(
            object_link(obj), u'<a href="http://nohost/plone/folder/view">Folder title &lt;script /&gt;</a>'
        )
        self.assertEqual(
            object_link(obj, escaped=False), u'<a href="http://nohost/plone/folder/view">Folder title <script /></a>'
        )

    def test_separate_images(self):
        # no image, content is returned as is
        text = "<p>My text.</p><p>My text.</p><p>My text.</p>"
        result = separate_images(text)
        self.assertEqual(text, result)
        # one image, nothing changed
        text = '<p><img src="http://plone/nohost/image1.png"></p>'
        result = separate_images(text)
        self.assertEqual(text, result)
        # one image with text, nothing changed
        text = '<p><img src="http://plone/nohost/image1.png">.</p>'
        result = separate_images(text)
        self.assertEqual(text, result)
        # one image with text, nothing changed
        text = '<p>Example : <img src="http://plone/nohost/image1.png"></p>'
        result = separate_images(text)
        self.assertEqual(text, result)
        # 2 images with text, nothing changed
        text = (
            '<p>Example : <img src="http://plone/nohost/image1.png">' '<img src="http://plone/nohost/image2.png"></p>'
        )
        result = separate_images(text)
        self.assertEqual(text, result)

        # <p> containg other tags than <img>, nothing changed
        text = (
            '<p><img src="http://plone/nohost/image1.png">'
            '<img src="http://plone/nohost/image2.png"><span>Text</span></p>'
        )
        result = separate_images(text)
        self.assertEqual(text, result)

        # now working examples
        # 2 images
        text = '<p><img src="http://plone/nohost/image1.png">' '<img src="http://plone/nohost/image2.png"></p>'
        result = separate_images(text)
        self.assertEqual(
            result,
            '<p><img src="http://plone/nohost/image1.png"></p>' '<p><img src="http://plone/nohost/image2.png"></p>',
        )
        # 3 images
        text = (
            '<p><img src="http://plone/nohost/image1.png">'
            '<img src="http://plone/nohost/image2.png">'
            '<img src="http://plone/nohost/image3.png"></p>'
        )
        result = separate_images(text)
        self.assertEqual(
            result,
            '<p><img src="http://plone/nohost/image1.png"></p>'
            '<p><img src="http://plone/nohost/image2.png"></p>'
            '<p><img src="http://plone/nohost/image3.png"></p>',
        )
        # complex case
        text = (
            "<p>My text and so on...</p>"
            '<p><img src="http://plone/nohost/image1.png"></p>'
            '<p><img src="http://plone/nohost/image2.png">'
            '<img src="http://plone/nohost/image3.png"></p>'
            "<p>Some other text...</p>"
            '<p><img src="http://plone/nohost/image4.png"></p>'
        )
        result = separate_images(text)
        self.assertEqual(
            result,
            "<p>My text and so on...</p>"
            '<p><img src="http://plone/nohost/image1.png"></p>'
            '<p><img src="http://plone/nohost/image2.png"></p>'
            '<p><img src="http://plone/nohost/image3.png"></p>'
            "<p>Some other text...</p>"
            '<p><img src="http://plone/nohost/image4.png"></p>',
        )
        # very complex case
        text = (
            "<p>My text and so on...</p>"
            '<p><img src="http://plone/nohost/image1.png"></p>'
            '<div><img src="http://plone/nohost/image2.png">'
            '<img src="http://plone/nohost/image3.png"></div>'
            "<p>Some other text...</p>"
            '<p><img src="http://plone/nohost/image4.png">'
            '<img src="http://plone/nohost/image5.png"></p>'
            '<p><img src="http://plone/nohost/image6.png"></p>'
            '<table><tr><td><p><img src="http://plone/nohost/image7.png">'
            '<img src="http://plone/nohost/image8.png"></p></td></tr></table>'
        )
        result = separate_images(text)
        self.assertEqual(
            result,
            "<p>My text and so on...</p>"
            '<p><img src="http://plone/nohost/image1.png"></p>'
            '<div><img src="http://plone/nohost/image2.png"></div>'
            '<div><img src="http://plone/nohost/image3.png"></div>'
            "<p>Some other text...</p>"
            '<p><img src="http://plone/nohost/image4.png"></p>'
            '<p><img src="http://plone/nohost/image5.png"></p>'
            '<p><img src="http://plone/nohost/image6.png"></p>'
            '<table><tr><td><p><img src="http://plone/nohost/image7.png">'
            '<img src="http://plone/nohost/image8.png"></p></td></tr></table>',
        )
        # <br> are ignored
        text = (
            '<p><img src="http://plone/nohost/image1.png"><br><br />' '<img src="http://plone/nohost/image2.png"></p>'
        )
        result = separate_images(text)
        self.assertEqual(
            result,
            '<p><img src="http://plone/nohost/image1.png"><br><br></p>'
            '<p><img src="http://plone/nohost/image2.png"></p>',
        )
        # blanks are ignored
        text = (
            '<p><img src="http://plone/nohost/image1.png">&nbsp; &nbsp;'
            '<img src="http://plone/nohost/image2.png"></p>'
        )
        result = separate_images(text)
        if six.PY2:
            self.assertEqual(
                result,
                '<p><img src="http://plone/nohost/image1.png">\xc2\xa0 \xc2\xa0</p>'
                '<p><img src="http://plone/nohost/image2.png"></p>',
            )
        else:
            self.assertEqual(
                result,
                '<p><img src="http://plone/nohost/image1.png">\xa0 \xa0</p>'
                '<p><img src="http://plone/nohost/image2.png"></p>',
            )
        # blanks and <br> are ignored as well
        text = (
            '<p><img src="http://plone/nohost/image1.png">&nbsp; &nbsp;<br>'
            '<img src="http://plone/nohost/image2.png"></p>'
        )
        result = separate_images(text)
        if six.PY2:
            self.assertEqual(
                result,
                '<p><img src="http://plone/nohost/image1.png">\xc2\xa0 \xc2\xa0<br></p>'
                '<p><img src="http://plone/nohost/image2.png"></p>',
            )
        else:
            self.assertEqual(
                result,
                '<p><img src="http://plone/nohost/image1.png">\xa0 \xa0<br></p>'
                '<p><img src="http://plone/nohost/image2.png"></p>',
            )

    def test_replace_content(self):
        text = (
            '<p>Text <span class="to-hide">hidden</span> some other text.</p>'
            '<p>Text <span class="to-hide">hidden <strong>hidden</strong> hidden</span> some other text.</p>'
            '<p><span class="to-hide">hidden</span> some other text.</p>'
            '<p><span class="to-hide">hidden</span></p>'
            '<p><span class="to-hide">hidden</span></p>'
            '<table><tr><td>Text <span class="to-hide">hidden</span></td>'
            "<td>Text not hidden</td></tr></table>"
        )
        expected = (
            '<p>Text <span class="to-hide"></span> some other text.</p>'
            '<p>Text <span class="to-hide"></span> some other text.</p>'
            '<p><span class="to-hide"></span> some other text.</p>'
            '<p><span class="to-hide"></span></p>'
            '<p><span class="to-hide"></span></p>'
            '<table><tr><td>Text <span class="to-hide"></span></td>'
            "<td>Text not hidden</td></tr></table>"
        )
        expected_new_content = (
            '<p>Text <span class="to-hide">replaced</span> some other text.</p>'
            '<p>Text <span class="to-hide">replaced</span> some other text.</p>'
            '<p><span class="to-hide">replaced</span> some other text.</p>'
            '<p><span class="to-hide">replaced</span></p>'
            '<p><span class="to-hide">replaced</span></p>'
            '<table><tr><td>Text <span class="to-hide">replaced</span></td>'
            "<td>Text not hidden</td></tr></table>"
        )
        res = replace_content(text, css_class="to-hide")
        self.assertEqual(res, expected)
        res = replace_content(text, css_class="to-hide", new_content=u"replaced")
        self.assertEqual(res, expected_new_content)
        text_link = '<p>Text <span class="to-hide">hidden <strong>hidden</strong></span></p>'
        expected_link = (
            '<p>Text <span class="to-hide">' '<a href="https://python.org" title="Explanation">replaced</a></span></p>'
        )
        res = replace_content(
            text_link,
            css_class="to-hide",
            new_content=u"replaced",
            new_content_link={"href": "https://python.org", "title": u"Explanation"},
        )
        self.assertEqual(res, expected_link)
        # every sub content is removed
        text_with_sub = (
            '<p>Text <span class="to-hide"><span class="highlight">hidden</span>' "</span> some other text.</p>"
        )
        expected_text_with_sub = '<p>Text <span class="to-hide">replaced</span> some other text.</p>'
        res = replace_content(text_with_sub, css_class="to-hide", new_content=u"replaced")
        self.assertEqual(res, expected_text_with_sub)

    def test_is_html(self):
        self.assertFalse(is_html(None))
        self.assertFalse(is_html(""))
        self.assertFalse(is_html("None"))
        self.assertTrue(is_html("<p>My text</p>"))
        self.assertTrue(is_html("<p>My text <span>text 2</span>.</p>"))

    def test_unescape_html(self):
        self.assertIsNone(unescape_html(None))
        self.assertEqual(unescape_html(""), "")
        self.assertEqual(unescape_html(u""), u"")
        self.assertEqual(unescape_html("<p>Activit&#233; scolaire</p>"), "<p>Activité scolaire</p>")
        self.assertEqual(unescape_html(u"<p>Activit&#233; scolaire</p>"), u"<p>Activité scolaire</p>")



# -*- coding: utf-8 -*-

from plone import api
from plone.api.exc import InvalidParameterError
from plone.dexterity.fti import DexterityFTI
from Products.GenericSetup.interfaces import IBody
from Products.GenericSetup.utils import importObjects
from zope.component import queryMultiAdapter

import logging


logger = logging.getLogger("imio.helpers.setup")


def load_type_from_package(type_name, profile_id, purge_actions=False):
    """Loads a portal_type from his xml definition.
    :param type_name: portal_type id
    :param profile_id: package profile id
    :param purge_actions: empties type actions
    :return: status as boolean
    """
    types_tool = api.portal.get_tool("portal_types")
    portal_type = types_tool.get(type_name)
    if portal_type is None:
        logger.error("Cannot find '{}' portal_type name in portal".format(type_name))
        return False
    ps_tool = api.portal.get_tool("portal_setup")
    try:
        context = ps_tool._getImportContext(profile_id, True)
    except KeyError:
        logger.error("Cannot find '{}' profile id".format(profile_id))
        return False

    # special case for DX FTI, _should_purge is set to False or it fails when purging
    if isinstance(portal_type, DexterityFTI):
        context._should_purge = False

    if purge_actions:
        # remove actions so it is reloaded in correct order
        portal_type._actions = ()

    # ps_tool.applyContext(context)  # necessary ?
    importObjects(portal_type, "types/", context)
    if portal_type._p_changed is False:
        logger.error("Could not update '{}' using profile '{}'".format(type_name, profile_id))
        return False
    return True


def load_workflow_from_package(wkf_name, profile_id):
    """Loads a workflow from his xml definition.
    :param wkf_name: workflow id
    :param profile_id: package profile id
    :return: status as boolean
    """
    wkf_tool = api.portal.get_tool("portal_workflow")
    wkf_obj = wkf_tool.get(wkf_name)
    if wkf_obj is None:
        logger.error("Cannot find '{}' workflow name in portal".format(wkf_name))
        return False
    ps_tool = api.portal.get_tool("portal_setup")
    try:
        context = ps_tool._getImportContext(profile_id, True)
    except KeyError:
        logger.error("Cannot find '{}' profile id".format(profile_id))
        return False
    # ps_tool.applyContext(context)  # necessary ?
    importObjects(wkf_obj, "workflows/", context)
    if wkf_obj._p_changed is False:
        logger.error("Could not update '{}' using profile '{}'".format(wkf_name, profile_id))
        return False
    return True


def load_xml_tool_only_from_package(tool_name, profile_id):
    """Loads a tool from his xml definition.
    :param tool_name: tool id
    :param profile_id: package profile id
    :return: status as boolean
    """
    raise NotImplementedError
    try:
        tool = api.portal.get_tool(tool_name)
    except InvalidParameterError:
        logger.error("Cannot find '{}' tool name in portal".format(tool_name))
        return False
    ps_tool = api.portal.get_tool("portal_setup")
    try:
        context = ps_tool._getImportContext(profile_id, False)  # do not purge !
    except KeyError:
        logger.error("Cannot find '{}' profile id".format(profile_id))
        return False
    # ps_tool.applyContext(context)  # necessary ?
    importer = queryMultiAdapter((tool, context), IBody)
    path = tool_name.replace(" ", "_")
    __traceback_info__ = path
    if importer:
        if importer.name:
            path = importer.name
        filename = "%s%s" % (path, importer.suffix)
        body = context.readDataFile(filename)
        if body is not None:
            importer.filename = filename  # for error reporting
            importer.body = body
    if tool._p_changed is False:
        logger.error("Could not update '{}' using profile '{}'".format(tool_name, profile_id))
        return False
    return True


def remove_gs_step(step_id, registry="_import_registry"):
    """Remove a step from a generic setup registry.
    :param step_id: import step id
    :param registry: registry name (default: _import_registry)
    :return: status as boolean
    """
    ps_tool = api.portal.get_tool("portal_setup")
    if not hasattr(ps_tool, registry):
        logger.error("Cannot find '{}' registry in portal_setup".format(registry))
        return False
    registry = getattr(ps_tool, registry)
    if step_id in registry.listSteps():
        registry.unregisterStep(step_id)
        ps_tool._p_changed = True
        return True
    return False



# -*- coding: utf-8 -*-

from imio.helpers import browser
from imio.helpers import path

import unittest


class TestPath(unittest.TestCase):
    def test_path_to_package(self):
        self.assertTrue(path.path_to_package(path).endswith("/src/imio/helpers/"))
        self.assertTrue(
            path.path_to_package(browser, filepart="static/listings.js").endswith(
                "/src/imio/helpers/browser/static/listings.js"
            )
        )



# -*- coding: utf-8 -*-
from plone import api


def uninstall_product(context):
    if context.readDataFile("imio.helpers_uninstall.txt") is None:
        return

    setup_tool = api.portal.get_tool("portal_setup")
    setup_tool.unsetLastVersionForProfile("imio.helpers:install-base")



# -*- coding: utf-8 -*-
import os


def path_to_package(package, filepart=""):
    """Return the absolute path to p_filepart stored in p_package."""
    return os.path.join(os.path.dirname(package.__file__), filepart)



# -*- coding: utf-8 -*-

from datetime import date
from datetime import datetime
from imio.helpers.cache import get_cachekey_volatile
from plone import api
from plone.memoize import ram as pmram
from Products.PlonePAS.plugins.role import GroupAwareRoleManager
from Products.PlonePAS.tools.groups import GroupsTool
from Products.PluggableAuthService.PluggableAuthService import PluggableAuthService
from zope.globalrequest import getRequest
from zope.i18nmessageid import MessageFactory

import logging
import os


_ = MessageFactory("imio.helpers")
logger = logging.getLogger("imio.helpers")

HAS_PLONE_4 = api.env.plone_version().startswith("4")
HAS_PLONE_5 = api.env.plone_version().startswith("5")
HAS_PLONE_5_1 = api.env.plone_version() > "5.1"
HAS_PLONE_5_2 = api.env.plone_version() > "5.2"
HAS_PLONE_5_AND_MORE = int(api.env.plone_version()[0]) >= 5
HAS_PLONE_6_AND_MORE = int(api.env.plone_version()[0]) >= 6
PLONE_MAJOR_VERSION = int(api.env.plone_version()[0])

EMPTY_TITLE = "No value"
EMPTY_STRING = "__empty_string__"
EMPTY_DATE = date(1950, 1, 1)
EMPTY_DATETIME = datetime(1950, 1, 1, 12, 0)


def GroupsTool__getGroupsForPrincipal_cachekey(method, self, principal):
    req = getRequest()
    if req is None:
        raise pmram.DontCache
    date = get_cachekey_volatile("_users_groups_value")
    return date, principal and principal.getId()


def GroupAwareRoleManager__getRolesForPrincipal_cachekey(method, self, principal, request=None):
    req = request or getRequest()
    # if req is None:
    #     raise pmram.DontCache
    date = get_cachekey_volatile("_users_groups_value")
    return (
        date,
        principal and principal.getId(),
        repr(req),
        req and (req.get("__ignore_direct_roles__", False), req.get("__ignore_group_roles__", False)) or (None, None),
    )


def PluggableAuthService__getGroupsForPrincipal_cachekey(method, self, principal, request=None, **kwargs):
    req = request or getRequest()
    if req is None:
        raise pmram.DontCache
    try:
        date = get_cachekey_volatile("_users_groups_value")
    except api.portal.CannotGetPortalError:
        raise pmram.DontCache
    return date, principal and principal.getId()


def PluggableAuthService__findUser_cachekey(method, self, plugins, user_id, name=None, request=None):
    req = request or getRequest()
    if req is None:
        raise pmram.DontCache
    try:
        date = get_cachekey_volatile("_users_groups_value")
    except api.portal.CannotGetPortalError:
        raise pmram.DontCache
    return date, repr(plugins), user_id, name, str(req and req._debug or "")


def PluggableAuthService__verifyUser_cachekey(method, self, plugins, user_id=None, login=None):
    req = getRequest()
    if req is None:
        raise pmram.DontCache

    date = get_cachekey_volatile("_users_groups_value")
    return date, repr(plugins), user_id, login


def GroupsTool_getGroupById_cachekey(method, self, group_id):
    req = getRequest()
    if req is None:
        raise pmram.DontCache

    date = get_cachekey_volatile("_users_groups_value")
    return date, group_id


if os.getenv("decorate_acl_methods", "Nope") in ("True", "true"):
    logger.info("DECORATING various acl related methods with cache")
    decorator = pmram.cache(GroupsTool__getGroupsForPrincipal_cachekey)
    GroupsTool.getGroupsForPrincipal = decorator(GroupsTool.getGroupsForPrincipal)
    decorator = pmram.cache(GroupAwareRoleManager__getRolesForPrincipal_cachekey)
    GroupAwareRoleManager.getRolesForPrincipal = decorator(GroupAwareRoleManager.getRolesForPrincipal)
    decorator = pmram.cache(PluggableAuthService__getGroupsForPrincipal_cachekey)
    PluggableAuthService._getGroupsForPrincipal = decorator(PluggableAuthService._getGroupsForPrincipal)
    # decorator = pmram.cache(PluggableAuthService__findUser_cachekey)
    # PluggableAuthService._findUser = decorator(PluggableAuthService._findUser)
    # decorator = pmram.cache(PluggableAuthService__verifyUser_cachekey)
    # PluggableAuthService._verifyUser = decorator(PluggableAuthService._verifyUser)
    decorator = pmram.cache(GroupsTool_getGroupById_cachekey)
    GroupsTool.getGroupById = decorator(GroupsTool.getGroupById)



# -*- coding: utf-8 -*-

from ZPublisher.Converters import type_converters

import json


try:
    from cgi import escape  # Python 2.x
except ImportError:
    from html import escape


# create type converter for :json
if "json" not in type_converters:

    def field2json(v):
        try:
            v = json.loads(v)
        except ValueError:
            raise ValueError("Invalid json " + escape(repr(v)))
        return v

    type_converters["json"] = field2json



# -*- coding: utf-8 -*-

from imio.helpers.cache import invalidate_cachekey_volatile_for


def onPrincipalCreated(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onPrincipalDeleted(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onPrincipalModified(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onPrincipalAddedToGroup(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onPrincipalRemovedFromGroup(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onGroupCreated(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)


def onGroupDeleted(event):
    """Invalidate the cachekey that manages users/groups associations."""
    invalidate_cachekey_volatile_for("_users_groups_value", get_again=True)



# -*- coding: utf-8 -*-
from imio.helpers import HAS_PLONE_6_AND_MORE
from imio.helpers.setup import load_type_from_package
from imio.helpers.setup import load_workflow_from_package
from imio.helpers.setup import load_xml_tool_only_from_package
from imio.helpers.setup import remove_gs_step
from imio.helpers.testing import IntegrationTestCase
from plone import api

import six
import unittest


class TestSetupModule(IntegrationTestCase):
    """
    Test all helper methods of setup module.
    """

    def setUp(self):
        self.portal = self.layer["portal"]
        try:
            from Products.CMFPlone.utils import get_installer

            self.installer = get_installer(self.portal)
        except ImportError:
            self.installer = None

    def test_load_type_from_package(self):
        if six.PY2:
            package = "profile-Products.CMFPlone:plone"
        else:
            package = "plone.app.contenttypes:default"
        types_tool = api.portal.get_tool("portal_types")
        portal_type = types_tool.get("File")
        self.assertTrue(portal_type.filter_content_types)
        portal_type.filter_content_types = False
        self.assertFalse(portal_type.filter_content_types)
        self.assertTrue(load_type_from_package("File", package))
        portal_type = types_tool.get("File")
        self.assertTrue(portal_type.filter_content_types)
        # not found portal_type
        self.assertFalse(load_type_from_package("Folder2", package))
        # not found profile_id
        self.assertFalse(load_type_from_package("Folder", package + "2"))
        # type not managed by given profile_id
        self.assertFalse(load_type_from_package("testingtype", package))
        # reimport a Dexterity fti
        self.assertTrue(load_type_from_package("testingtype", "profile-imio.helpers:testing"))
        # with purge_actions=True
        self.assertTrue(load_type_from_package("testingtype", "profile-imio.helpers:testing", purge_actions=True))

    def test_load_workflow_from_package(self):
        wkf_tool = api.portal.get_tool("portal_workflow")
        wkf_obj = wkf_tool.get("intranet_workflow")
        states = wkf_obj.states
        self.assertIn("internal", states)
        states.deleteStates(["internal"])
        self.assertNotIn("internal", states)
        self.assertTrue(load_workflow_from_package("intranet_workflow", "profile-Products.CMFPlone:plone"))
        wkf_obj = wkf_tool.get("intranet_workflow")
        states = wkf_obj.states
        self.assertIn("internal", states)
        # not found WF
        self.assertFalse(load_workflow_from_package("intranet_workflow2", "profile-Products.CMFPlone:plone"))
        # not found profile_id
        self.assertFalse(load_workflow_from_package("intranet_workflow", "profile-Products.CMFPlone:plone2"))
        # WF not managed by given profile_id
        self.assertFalse(load_workflow_from_package("comment_review_workflow", "profile-Products.CMFPlone:plone"))

    @unittest.skip("Not working")
    def test_load_xml_tool_only_from_package(self):
        wkf_tool = api.portal.get_tool("portal_workflow")
        wkf_obj = wkf_tool.get("intranet_workflow")
        states = wkf_obj.states
        self.assertIn("internal", states)
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Image"), ())
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Document"), ("intranet_workflow",))
        # some changes
        states.deleteStates(["internal"])
        self.assertNotIn("internal", states)
        wkf_tool.setChainForPortalTypes(("Image",), ("one_state_workflow",))
        wkf_tool.setChainForPortalTypes(("Document",), ("one_state_workflow",))
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Image"), ("one_state_workflow",))
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Document"), ("one_state_workflow",))
        # we reload default config
        self.assertTrue(load_xml_tool_only_from_package("portal_workflow", "profile-Products.CMFPlone:plone"))
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Image"), ())  # chain is reset
        self.assertTupleEqual(wkf_tool.getChainForPortalType("Document"), ("one_state_workflow",))  # not in xml
        wkf_obj = wkf_tool.get("intranet_workflow")
        states = wkf_obj.states
        self.assertNotIn("internal", states)  # be sure it's not recursive
        # not found tool
        self.assertFalse(load_xml_tool_only_from_package("portal_workflow2", "profile-Products.CMFPlone:plone"))
        # not found profile_id
        self.assertFalse(load_xml_tool_only_from_package("portal_workflow", "profile-Products.CMFPlone:plone2"))

    def test_remove_gs_step(self):
        ps_tool = api.portal.get_tool("portal_setup")
        step_id = u"mockmailhost-various"
        self.assertIn(step_id, ps_tool.getSortedImportSteps())
        self.assertTrue(remove_gs_step(step_id))
        self.assertNotIn(step_id, ps_tool.getSortedImportSteps())

    def test_product_installed(self):
        """Test if imio.helpers is installed with portal_quickinstaller."""
        if not HAS_PLONE_6_AND_MORE:
            return
        self.assertTrue(self.installer.is_product_installed("imio.helpers"))

    def test_uninstall(self):
        """Test if imio.helpers is cleanly uninstalled."""
        if not HAS_PLONE_6_AND_MORE:
            return
        self.installer.uninstall_product("imio.helpers")
        self.assertFalse(self.installer.is_product_installed("imio.helpers"))



# -*- coding: utf-8 -*-

from imio.helpers import HAS_PLONE_6_AND_MORE
from imio.helpers.testing import IntegrationTestCase
from unittest import TestLoader
from unittest import TestSuite
from ZPublisher.tests.testHTTPRequest import HTTPRequestTests

import html
import imio.helpers.converters  # noqa


class TestConverters(IntegrationTestCase, HTTPRequestTests):
    def test_imiohelpers_json_converter(self):
        inputs = (
            ("data:json", '{"key1": "value1", "key2": "value2"}'),
            ("data2:json", '{"key3": "value3", "key4": "value4"}'),
        )
        req = self._processInputs(inputs)
        self.assertEqual(
            req.form,
            {"data": {u"key2": u"value2", u"key1": u"value1"}, "data2": {u"key3": u"value3", u"key4": u"value4"}},
        )

    def test_imiohelpers_json_converter_wrong_format(self):
        # missing " after value1
        inputs = (("data:json", '{"key1": "value1, "key2": "value2"}'),)
        with self.assertRaises(ValueError) as cm:
            self._processInputs(inputs)
        error_message = 'Invalid json \'{"key1": "value1, "key2": "value2"}\''
        if HAS_PLONE_6_AND_MORE:
            self.assertEqual(html.unescape(cm.exception.args[0]), error_message)
        else:
            self.assertEqual(cm.exception.message, error_message)


def test_suite():
    suite = TestSuite()
    for name in TestLoader().getTestCaseNames(TestConverters):
        # change prefix to avoid executing every tests of HTTPRequestTests
        if name.startswith("test_imiohelpers_"):
            suite.addTest(TestConverters(name))
    return suite



# -*- coding: utf-8 -*-

import os


def get_file_data(filename):
    file_path = os.path.join(os.path.dirname(__file__), filename)
    fd = open(file_path, "rb")
    data = fd.read()
    fd.close()
    return data


def restricted_method(self, *args, **kwargs):
    print("%s : Archetype is required for this test" % self.__name__)
    # raise Exception(f"{module_name} is not installed, can't access this method.")


def require_module(module_name):
    try:
        __import__(module_name)
        return lambda func: func
    except ImportError:
        return restricted_method


def unrequire_module(module_name):
    try:
        __import__(module_name)
        return restricted_method
    except ImportError:
        return lambda func: func



# -*- coding: utf-8 -*-

import six
import subprocess


def generate_barcode(data, executable='zint', barcode=92, scale=2, filetype='PNG', extra_args=None):
    """Generate a barcode with zint in BytesIO and return it
    :param data: barcode content string
    :param executable: executable used to generate barcode
    :param barcode: barcode type number
    :param scale: scale of generated output
    :param filetype: image
    :param extra_args: additional arguments for zint, for example ["--secure=2"]
    :return: BytesIO handler
    """
    output = six.BytesIO()

    command = [
        executable,
        # '--directpng',  # no more used in version 2.6.0, and replaced by 2 following parameters
        "--direct",
        "--filetype={0}".format(filetype),
        "--barcode={0}".format(barcode),
        "--scale={0}".format(scale),
    ]
    if extra_args:
        command += extra_args
    # Data must be the last argument
    command.append("--data={0}".format(data))

    process = subprocess.Popen(command, stdout=subprocess.PIPE)
    output.write(process.stdout.read())
    process.stdout.close()
    process.wait()

    output.seek(0)
    return output



# -*- coding: utf-8 -*-

from DateTime import DateTime
from plone import api
from six.moves.cPickle import dumps
from time import time
from zope.globalrequest import getRequest
from zope.ramcache.ram import caches
from zope.ramcache.ram import RAMCache
from zope.ramcache.ram import Storage
from zope.ramcache.ram import writelock

import six


class IMIORAMCache(RAMCache):
    """ """

    def _getStorage(self):
        """Finds or creates a storage object."""
        cacheId = self._cacheId
        with writelock:
            if cacheId not in caches:
                # imio.helpers, only changed here, use IMIOStorage instead Storage
                caches[cacheId] = IMIOStorage(self.maxEntries, self.maxAge, self.cleanupInterval)
            return caches[cacheId]


class IMIOStorage(Storage):
    """ """

    def getEntry(self, ob, key):
        if self.lastCleanup <= time() - self.cleanupInterval:
            self.cleanup()

        try:
            data = self._data[ob][key]
        except KeyError:
            if ob not in self._misses:
                self._misses[ob] = 0
            self._misses[ob] += 1
            raise
        else:
            if isinstance(data, list):
                # [data, ctime, access count]
                data[2] += 1  # increment access count
                # XXX begin change by imio.helpers, update timestamp
                timestamp = time()
                data[1] = timestamp
                # XXX end change by PM
                return data[0]
            else:
                data.access_count += 1  # increment access count
                # XXX begin change by imio.helpers, update timestamp
                timestamp = time()
                data.ctime = timestamp
                # XXX end change by PM
                return data.value

    def getStatistics(self):
        objects = list(self._data.keys())
        if six.PY3:
            # prevent AttributeError: 'dict_keys' object has no attribute 'sort'
            objects = list(objects)
        objects.sort()
        result = []

        for ob in objects:
            try:
                size = len(dumps(self._data[ob]))
            except Exception as exc:
                size = 0
                api.portal.show_message(
                    'Could not compute size for "%s", original exception was "%s"' % (repr(ob), repr(exc)),
                    request=getRequest(),
                )

            hits = 0
            older_date = None
            for entry in six.itervalues(self._data[ob]):
                if isinstance(entry, list):
                    # [data, ctime, access count]
                    hits += entry[2]
                    older_date = older_date and min(older_date, entry[1]) or entry[1]
                else:
                    hits += entry.access_count
                    older_date = older_date and min(older_date, entry.ctime) or entry.ctime

            result.append(
                {
                    "path": ob,
                    "hits": hits,
                    "misses": self._misses.get(ob, 0),
                    "size": size,
                    "entries": len(self._data[ob]),
                    "older_date": older_date and DateTime(older_date),
                }
            )
        return tuple(result)


# instantiate our RAM Cache
imio_global_cache = IMIORAMCache()



# -*- coding: utf-8 -*-

from imio.helpers.testing import IntegrationTestCase
from imio.helpers.tests.utils import require_module
from plone import api
from plone.app.textfield.value import RichTextValue


class TestAppyPod(IntegrationTestCase):
    @require_module("Products.Archetypes")
    def test_load_appy_pod_sample_archetypes(self):
        """Test for Archetypes."""
        docId = self.portal.invokeFactory("Document", id="doc", title="Original title")
        doc = getattr(self.portal, docId)
        view = doc.restrictedTraverse("@@appy_pod_sample")
        # does not break if called on an unexisting field or
        # not RichText field, just do nothing and display a portal_message
        view(field_name="title")
        self.assertEqual(doc.Title(), "Original title")
        view(field_name="unexisting_field_name")
        # call it on a real RichText field_name
        self.assertEqual(doc.getText(), "")
        view(field_name="text")
        self.assertNotEquals(doc.getText(), "")

    def test_load_appy_pod_sample_dexterity(self):
        """Test for dexterity."""
        doc = api.content.create(container=self.portal, type="testingtype", id="testingtype", title="Original title")
        view = doc.restrictedTraverse("@@appy_pod_sample")
        # does not break if called on an unexisting field or
        # not RichText field, just do nothing and display a portal_message
        view(field_name="title")
        self.assertEqual(doc.Title(), "Original title")
        view(field_name="unexisting_field_name")
        # call it on a real RichText field_name
        self.assertEqual(doc.text, None)
        view(field_name="text")
        self.assertTrue(isinstance(doc.text, RichTextValue))



# -*- coding: utf-8 -*-

from App.config import _config
from imio.helpers.security import check_zope_admin
from imio.helpers.security import fplog
from imio.helpers.security import generate_password
from imio.helpers.security import get_environment
from imio.helpers.security import get_user_from_criteria
from imio.helpers.security import get_zope_root
from imio.helpers.security import is_develop_environment
from imio.helpers.security import set_site_from_package_config
from imio.helpers.security import setup_logger
from imio.helpers.testing import IntegrationTestCase
from OFS.Application import Application
from plone import api
from plone.app.testing import login
from Products.CMFCore.permissions import ManagePortal
from Products.CMFCore.utils import _checkPermission
from Products.CMFPlone.Portal import PloneSite

import os
import re


class TestSecurityModule(IntegrationTestCase):
    """
    Test all helper methods of security module.
    """

    def setUp(self):
        if "IS_DEV_ENV" in list(os.environ.keys()):
            del os.environ["IS_DEV_ENV"]
        if "ENV" in list(os.environ.keys()):
            del os.environ["ENV"]
        self.portal = self.layer["portal"]

    def test_fplog(self):
        fplog("test_action", "test_extras")

    def test_is_develop_environment(self):
        os.environ["IS_DEV_ENV"] = "false"
        self.assertFalse(is_develop_environment())
        os.environ["IS_DEV_ENV"] = "true"
        self.assertTrue(is_develop_environment())

    def test_get_environment(self):
        os.environ["ENV"] = "staging"
        self.assertFalse(is_develop_environment())
        self.assertEqual(get_environment(), "staging")
        os.environ["ENV"] = "dev"
        self.assertEqual(get_environment(), "dev")
        self.assertTrue(is_develop_environment())

    def test_generate_password(self):
        pwd = generate_password(length=10, digits=1, upper=1, lower=1, special=1, readable=False)
        self.assertEqual(len(pwd), 10)
        self.assertIsNotNone(re.search(r"\d", pwd))
        self.assertIsNotNone(re.search(r"[A-Z]", pwd))
        self.assertIsNotNone(re.search(r"[a-z]", pwd))
        self.assertIsNotNone(re.search(r"[!#$%&*+<=>?@-]", pwd))

    def test_get_user_from_criteria(self):
        api.user.create("elementary@mail.be", "ssmith", "secret1234", properties={"fullname": "Stéphan Smith"})
        api.user.create("teacher@school.be", "teacher", "secret1234", properties={"fullname": "Paul Smith"})
        self.assertEqual(len(get_user_from_criteria(self.portal, email="unknown")), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname="unknown")), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid="unknown")), 0)
        self.assertEqual(len(get_user_from_criteria(self.portal, email="mentary")), 1)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname="Stéph")), 1)
        self.assertEqual(len(get_user_from_criteria(self.portal, email=".be")), 2)
        self.assertEqual(len(get_user_from_criteria(self.portal, email="")), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname="Smith")), 2)
        self.assertEqual(len(get_user_from_criteria(self.portal, fullname="")), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid="")), 3)
        self.assertEqual(len(get_user_from_criteria(self.portal, userid="teach")), 1)

    def test_setup_logger(self):
        # just call it to check that it is not broken
        self.assertIsNone(setup_logger())

    def test_get_zope_root(self):
        app = get_zope_root()
        self.assertIsInstance(app, Application)

    def test_set_site_from_package_config(self):
        # no config
        self.assertIsNone(set_site_from_package_config("imio.helpers"))
        # we add config
        _config.product_config = {"imio.helpers": {"plone-path": "plone"}}
        site = set_site_from_package_config("imio.helpers")
        self.assertIsInstance(site, PloneSite)

    def test_check_zope_admin(self):
        # create and login as siteadmin
        self.portal.portal_membership.addMember("siteadmin", "siteadmin", ["Manager"], [])
        login(self.portal, "siteadmin")
        self.assertTrue(_checkPermission(ManagePortal, self.portal))
        self.assertFalse(check_zope_admin())
        # login as Zope admin
        login(self.portal.aq_parent, "admin")
        self.assertTrue(_checkPermission(ManagePortal, self.portal))
        self.assertTrue(check_zope_admin())



# -*- coding: utf-8 -*-

from plone import api
from Products.CMFCore.WorkflowCore import WorkflowException
from Products.CMFPlone.utils import safe_unicode
from zope.i18n import translate

import logging


logger = logging.getLogger("imio.helpers.wokflow")


def do_transitions(obj, transitions, warn=False):
    """
    Apply multiple transitions on obj
    """
    wkf_tool = api.portal.get_tool("portal_workflow")
    if not isinstance(transitions, (list, tuple)):
        transitions = [transitions]
    for tr in transitions:
        try:
            wkf_tool.doActionFor(obj, tr)
        except WorkflowException as exc:
            if warn:
                logger.warn("Cannot apply transition '%s' on obj '%s': '%s'" % (tr, obj, exc))


def get_state_infos(obj):
    """ """
    wkf_tool = api.portal.get_tool("portal_workflow")
    review_state = wkf_tool.getInfoFor(obj, "review_state")
    wf = wkf_tool.getWorkflowsFor(obj)[0]
    state = wf.states.get(review_state)
    state_title = state and state.title or review_state
    return {
        "state_name": review_state,
        "state_title": translate(safe_unicode(state_title), domain="plone", context=obj.REQUEST),
    }


def get_transitions(obj):
    """Return the ids of the available transitions as portal_workflow.getTransitionsFor
    will actually return a list of dict with various infos (id, title, name, ...) of
    the available transitions."""
    wf_tool = api.portal.get_tool("portal_workflow")
    return [tr["id"] for tr in wf_tool.getTransitionsFor(obj)]


def get_leading_transitions(wf, state_id, not_starting_with=None):
    """Returns transitions leading to a WF state.

    :param wf: wf object
    :param state_id: leading state id
    :param not_starting_with: transition id starting string
    :return: a list of WFtransitions
    """
    res = []
    for tr in wf.transitions.values():
        if tr.new_state_id == state_id and (not_starting_with is None or not tr.id.startswith(not_starting_with)):
            res.append(tr)
    return res


def remove_state_transitions(wf_name, state_id, remv_trs=[]):
    """Removes transitions defined on a state and possible duplicates

    :param wf_name: workflow string name
    :param state_id: state id
    :param remv_trs: list of transitions to remove
    :return: True if removed
    """
    wf_tool = api.portal.get_tool("portal_workflow")
    # if wf_name not in wf_tool:  # prefer an exception...
    wf = wf_tool[wf_name]
    if state_id not in wf.states:
        return False
    state = wf.states[state_id]
    orig_len = len(state.transitions)
    trs = []
    # keep right transitions and remove duplicates
    [trs.append(tr) for tr in list(state.transitions) if tr not in remv_trs and tr not in trs]
    if len(trs) != orig_len:
        state.transitions = tuple(trs)
        return True
    return False


def update_role_mappings_for(obj, wf=None):
    """Update role mappings regarding WF definition.
       If updated, the security indexes are reindexed.

    :param wf: the obj workflow
    :param obj: object to update role mappings for
    """
    if not wf:
        wf_tool = api.portal.get_tool("portal_workflow")
        wf = wf_tool.getWorkflowsFor(obj)[0]
    updated = wf.updateRoleMappingsFor(obj)
    if updated:
        obj.reindexObjectSecurity()
    return updated


def get_final_states(wf, ignored_transition_prefix="back", ignored_transition_ids=[]):
    """Return the final states of a workflow.
       Final states are states with only "back" transitions.
    :param wf: the obj workflow
    :param ignored_transition_prefix: the prefix of transitions to ignore,
    considered as "back" transitions
    :param ignored_transition_ids: some specific transition ids to ignore
    :return: True if removed
    """
    res = []
    for state in wf.states.values():
        not_final_transitions = [
            tr
            for tr in state.transitions
            if not tr.startswith(ignored_transition_prefix) and tr not in ignored_transition_ids
        ]
        # no other transitions left, then it is a final state
        if not not_final_transitions:
            res.append(state.id)
    return res



# -*- coding: utf-8 -*-

from plone import api
from plone.app.testing import login
from plone.app.testing import logout
from plone.app.testing.bbb import _createMemberarea
from plone.app.testing.helpers import setRoles
from plone.testing.z2 import Browser


class ImioTestHelpers:  # noqa
    """Helper class for tests.

    Assuming that self.layer, self.portal are defined in another class setUp...
    """

    def add_principal_to_groups(self, principal_id, group_ids):
        """We need to changeUser so getGroups is updated.

        :param principal_id: userid
        :param group_ids: group ids list
        """
        with api.env.adopt_roles(["Manager"]):
            for group_id in group_ids:
                self.portal.portal_groups.addPrincipalToGroup(principal_id, group_id)
            self.change_user(self.member.getId())

    def change_user(self, username):
        """Logs out currently logged user and logs in p_loginName."""
        logout()
        if username == "admin":
            login(self.layer["app"], username)
        else:
            login(self.portal, username)
        self.member = api.user.get_current()  # noqa
        self.portal.REQUEST["AUTHENTICATED_USER"] = self.member

    def create_user(self, username, pwd="humpf", roles=("Member",), groups=(), mb_area=False):
        """Creates a user named p_username with some p_roles.

        :param username: username
        :param pwd: password
        :param roles: roles
        :param groups: groups
        :param mb_area: bool to create member area
        :return: new user
        """
        with api.env.adopt_roles(["Manager"]):
            new_user = api.user.create(email="test@test.be", username=username, password=pwd, roles=[], properties={})
            setRoles(self.portal, username, roles)
            for group in groups:
                self.add_principal_to_groups(username, group)
            if mb_area:
                _createMemberarea(self.portal, username)
            return new_user

    def remove_principal_from_groups(self, principal_id, group_ids):
        """We need to changeUser so getGroups is updated.

        :param principal_id: userid
        :param group_ids: group ids list
        """
        with api.env.adopt_roles(["Manager"]):
            for group_id in group_ids:
                self.portal.portal_groups.removePrincipalFromGroup(principal_id, group_id)
            self.change_user(self.member.getId())


class BrowserTest(ImioTestHelpers):
    """
    Helper class for Browser tests.
    """

    def setUp(self):
        super(BrowserTest, self).setUp()
        self.browser = Browser(self.portal)
        self.browser.handleErrors = False

    def browser_login(self, user, password):
        login(self.portal, user)
        self.browser.open(self.portal.absolute_url() + "/logout")
        self.browser.open(self.portal.absolute_url() + "/login_form")
        self.browser.getControl(name="__ac_name").value = user
        self.browser.getControl(name="__ac_password").value = password
        self.browser.getControl(name="submit").click()



# -*- coding: utf-8 -*-
from email import encoders
from email.header import Header
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formataddr
from email.utils import getaddresses
from email.utils import parseaddr
from imio.helpers import _
from imio.pyutils.utils import safe_encode
from past.types import basestring
from plone import api
from Products.MailHost.MailHost import _encode_address_string
from six import iteritems
from smtplib import SMTPException
from unidecode import unidecode
from zope import schema
from zope.component import getMultiAdapter

import csv
import logging
import re
import socket


try:
    from plone.base.utils import safe_text as safe_text
except ImportError:
    from Products.CMFPlone.utils import safe_unicode as safe_text

try:
    from Products.CMFDefault.exceptions import EmailAddressInvalid
    from Products.CMFDefault.utils import checkEmailAddress
except ImportError:
    from Products.CMFPlone.RegistrationTool import checkEmailAddress
    from Products.CMFPlone.RegistrationTool import EmailAddressInvalid


logger = logging.getLogger("imio.helpers")
EMAIL_CHARSET = "utf-8"


def get_email_charset():
    """Character set to use for encoding the email."""
    portal = api.portal.get()
    return portal.getProperty("email_charset", EMAIL_CHARSET)


def get_mail_host(check=False):
    """Get the MailHost object.

    :param check: bool. Check if mailhost configuration is correct.
    :return: MailHost tool
    """
    if check:
        # check mailhost on 'smtp_host' and 'email_from_address'
        portal = api.portal.get()
        ctrl_overview = getMultiAdapter((portal, portal.REQUEST), name="overview-controlpanel")
        if not ctrl_overview.mailhost_warning():
            return api.portal.get_tool("MailHost")
    else:
        return api.portal.get_tool("MailHost")


# MAIN FUNCTIONS


def create_html_email(html, with_plain=True):
    """Returns an email message with an html body and optionally plain body.

    :param html: html body content
    :param with_plain: add plain text version (default=True)
    :return: MIMEMultipart instance
    :rtype: MIMEMultipart
    """
    charset = get_email_charset()
    html = safe_text(html, charset)

    # Don't check possible body content charset ('US-ASCII', get_email_charset(), 'UTF-8')
    body_charset = charset

    html = html.encode("ascii", "xmlcharrefreplace")
    html_part = MIMEText(html, "html", body_charset)

    # doing a multipart message content for text and html
    email_content = MIMEMultipart("alternative")
    email_content.epilogue = ""

    if with_plain:
        portal_transforms = api.portal.get_tool("portal_transforms")
        plain = portal_transforms.convert("html_to_text", html).getData()
        plain = safe_text(plain, charset)
        plain = plain.encode(body_charset, "replace")
        text_part = MIMEText(plain, "plain", body_charset)
        email_content.attach(text_part)

    # the email client will try to render the last part first !
    email_content.attach(html_part)

    # doing a multipart email that can "receive" attachments too
    eml = MIMEMultipart()
    eml.attach(email_content)

    return eml


@api.validation.at_least_one_of("filepath", "content")
def add_attachment(eml, filename, filepath=None, content=None):
    """Adds attachment to email instance.
    Must pass at least filepath or content.

    :param eml: email instance
    :param filename: file name string to store in header
    :param filepath: attachment file disk path
    :param content: attachment binary content
    """
    # Add as application/octet-stream: email client can usually download this automatically as attachment
    part = MIMEBase("application", "octet-stream")
    if filepath:
        # don't check file existence. Exception is preferred
        with open(filepath, "rb") as attachment:
            part.set_payload(attachment.read())
    if content:
        part.set_payload(content)

    encoders.encode_base64(part)
    # filename must be utf8 to be correctly displayed in gmail
    # in gmail with unicode: Content-Disposition: "attachment; filename=\"Réponse candidature ouvrier communal.odt\""
    # in gmail with utf8: Content-Disposition: attachment; filename="Réponse candidature ouvrier communal.odt" => OK
    part.add_header("Content-Disposition", "attachment", filename=safe_encode(filename))
    eml.attach(part)


def send_email(eml, subject, mfrom, mto, mcc=None, mbcc=None, replyto=None, immediate=False):
    """Sends an email with MailHost.

    :param eml: email instance
    :param subject: subject string
    :param mfrom: from string
    :param mto: to string or string list or (name, address) list
    :param mcc: cc string or string list or (name, address) list
    :param mbcc: bcc string or string list or (name, address) list
    :param replyto: reply-to string or string list or (name, address) list
    :param immediate: send email immediately without queuing (default False)
    :return: status
    :rtype: bool
    """
    mail_host = get_mail_host()
    if mail_host is None:
        logger.error("Cannot send email: mail host not well defined.")
        return False, "Mail host not well defined"

    charset = get_email_charset()
    subject = safe_text(subject, charset)
    # Convert all our address list inputs
    addrs = {
        "From": _email_list_to_string(mfrom, charset),
        "To": _email_list_to_string(mto, charset),
        "Cc": _email_list_to_string(mcc, charset),
        "reply-to": _email_list_to_string(replyto, charset),
    }
    mbcc = _email_list_to_string(mbcc, charset)
    # Add extra headers
    _addHeaders(
        eml, Subject=Header(subject, charset), **dict((k, Header(v, charset)) for k, v in iteritems(addrs) if v)
    )
    # Handle all recipients to add bcc: smtplib sends to all but a recipient not found in header is considered as bcc...
    # bug in python 3.10.12 with getaddresses and parameters like ["xx@yy.be", "", ""]
    all_recipients = [
        formataddr(pair) for pair in getaddresses([addr for addr in (addrs["To"], addrs["Cc"], mbcc) if addr])
    ]
    try:
        # send is protected by permission 'Use mailhost'
        # send remove from headers bcc but if in all_recipients, it is handled as bcc in email...
        mail_host.send(eml.as_string(), all_recipients, addrs["From"], immediate=immediate, charset=charset)
    except (socket.error, SMTPException) as e:
        logger.error(u"Cannot send email to '{}' with subject '{}': {}".format(mto, subject, e))
        return False, "Could not send email : {}".format(e)
    # sent successfully
    return True, ""


def _email_list_to_string(addr_list, charset="utf8"):
    """SecureMailHost's secureSend can take a list of email addresses
    in addition to a simple string.  We convert any email input into a
    properly encoded string."""
    if addr_list is None:
        return ""
    if isinstance(addr_list, basestring):
        addr_str = addr_list
    else:
        # if the list item is a string include it, otherwise assume it's a
        # (name, address) tuple and turn it into an RFC compliant string

        addresses = (isinstance(a, basestring) and a or formataddr(a) for a in addr_list)
        addr_str = ", ".join(str(_encode_address_string(a, charset)) for a in addresses)
    return addr_str


def _addHeaders(message, **kwargs):
    for key, value in iteritems(kwargs):
        del message[key]
        message[key] = value


class InvalidEmailAddressFormat(schema.ValidationError):
    """Exception for invalid address format with real name part."""

    __doc__ = _(u"Invalid email address format: 'real name <email>' or 'email (real name)'")


class InvalidEmailAddress(schema.ValidationError):
    """Exception for invalid address.
    `doc` method is used to return a dynamic message with real tested address.
    """

    def __init__(self, eml, *args, **kwargs):
        self.eml = eml

    def doc(self):
        return _(u"Invalid email address: '${eml}'", mapping={"eml": self.eml})


class InvalidEmailAddressCharacters(schema.ValidationError):
    """Exception for invalid realname.
    `doc` method is used to return a dynamic message with real tested address.
    """

    def __init__(self, value, *args, **kwargs):
        self.value = value

    def doc(self):
        return _(u"Realname: '${value}' cannot contain accented or special characters", mapping={"value": self.value})


def validate_email_address(value):
    """Email validator for email address with possible real name part.

    :param value: email value
    :return: tuple containing real name and address
    :rtype: tuple
    """
    if not value:
        return True
    eml = safe_text(value)
    realname = u""
    complex_form = any([b in eml and e in eml for b, e in ("<>", "()")])
    # Use parseaddr only when necessary to avoid correction like 'a @d.c' => 'a@d.c'
    # or to avoid bad split like 'a<a@d.c' => 'a@d.c'
    if complex_form:
        if '""' in eml:  # '"\\"' ends with '""' and must be corrected. We choose to un-double the doublequotes
            eml = re.sub(r'""([^"<]+)""\s*<(.*)>', r'"\1" <\2>', eml)
        realname, eml = parseaddr(eml)
        if not realname and not eml:
            raise InvalidEmailAddressFormat(value)
        if realname != unidecode(realname):
            raise InvalidEmailAddressCharacters(realname)
        # we check if the email has not been corrected by parseaddr, removing some characters like space
        if eml not in value:
            raise InvalidEmailAddress(u"'{}' => '{}' ?".format(value, eml))
    try:
        checkEmailAddress(eml)
    except EmailAddressInvalid:
        raise InvalidEmailAddress(eml)
    return realname, eml.lower()


def validate_email_addresses(value):
    """Email validator for email addresses (with possible real name part) separated by ','.

    :param value: email value
    :return: True
    """
    if not value:
        return True
    ret = []
    # we need to multiply doublequotes, otherwise they are removed by csv
    value = value.replace('"', '"""')
    # split addresses using csv
    for line in csv.reader([safe_encode(value)], delimiter=",", quotechar='"', skipinitialspace=True):
        for eml in line:
            ret.append(validate_email_address(eml))
    return ret



# -*- coding: utf-8 -*-
from imio.helpers.testing import IntegrationTestCase
from imio.helpers.transmogrifier import clean_value
from imio.helpers.transmogrifier import filter_keys
from imio.helpers.transmogrifier import get_correct_id
from imio.helpers.transmogrifier import get_correct_path
from imio.helpers.transmogrifier import get_main_path
from imio.helpers.transmogrifier import get_obj_from_path
from imio.helpers.transmogrifier import key_val
from imio.helpers.transmogrifier import pool_tuples
from imio.helpers.transmogrifier import relative_path
from imio.helpers.transmogrifier import split_text
from imio.helpers.transmogrifier import str_to_bool
from imio.helpers.transmogrifier import str_to_date

import datetime
import os
import six


# logger simulation for some tests
def logger(item, msg):
    item["errors"] += 1


class TestTesting(IntegrationTestCase):
    def assertRaisesRegex(self, *args, **kwargs):
        if six.PY3:
            return super().assertRaisesRegex(*args, **kwargs)
        return self.assertRaisesRegexp(*args, **kwargs)

    def test_clean_value(self):
        self.assertEqual(clean_value(None), None)
        self.assertEqual(clean_value(u"  strip  "), u"strip")
        self.assertEqual(clean_value(u" | strip  ", strip=u" |"), u"strip")
        self.assertEqual(clean_value(u" strip  \n  strip  "), u"strip\nstrip")
        self.assertEqual(clean_value(u" strip  \n  strip  ", osep=u"\r\n"), u"strip\r\nstrip")
        self.assertEqual(clean_value(u" strip  |  strip  ", isep=u"|"), u"strip|strip")
        self.assertEqual(clean_value(u"  \n strip  \n "), u"strip")
        self.assertEqual(clean_value(u' strip  \n "', strip=u' "'), u"strip")
        self.assertEqual(clean_value(u' strip  \n "', patterns=[(r'"', u"")]), u"strip")
        self.assertEqual(clean_value(u' strip  \n "\'"', patterns=[(r'^["\']+$', u"")]), u"strip")
        self.assertEqual(clean_value(u" strip  \n ", patterns=[]), u"strip")
        self.assertEqual(clean_value(u" strip  \n line2", patterns=[], osep=u"-"), u"strip-line2")
        self.assertEqual(
            clean_value(u" stri    p  \n line  2", patterns=[(r"\s{2,}", u" ")], osep=u"-"), u"stri p-line 2"
        )

    def test_get_correct_id(self):
        self.assertEqual(get_correct_id(self.portal, "abcde"), "abcde")
        self.assertIn("folder", self.portal.objectIds())
        self.assertEqual(get_correct_id(self.portal, "folder"), "folder-1")
        self.assertEqual(get_correct_id(self.portal, "folder", with_letter=True), "folder-a")
        self.assertEqual(get_correct_id(self.portal.folder, "abcde"), "abcde")
        self.portal.folder.invokeFactory("Document", id="abcde", title="Document")
        self.assertEqual(get_correct_id(self.portal.folder, "abcde"), "abcde-1")
        self.assertEqual(get_correct_id(self.portal.folder, "abcde", True), "abcde-a")
        self.portal.folder.invokeFactory("Document", id="abcde-1", title="Document")
        self.assertEqual(get_correct_id(self.portal.folder, "abcde"), "abcde-2")

    def test_get_correct_id_dic(self):
        lst = ["a"]
        self.assertEqual(get_correct_id(lst, "z"), "z")
        self.assertEqual(get_correct_id(lst, "a"), "a-1")
        self.assertEqual(get_correct_id(lst, "a", True), "a-a")
        lst.extend(["a-1", "a-a"])
        self.assertEqual(get_correct_id(lst, "a"), "a-2")
        self.assertEqual(get_correct_id(lst, "a", True), "a-b")
        lst.extend(["a-{}".format(lt) for lt in "bcdefghijklmnopqrstuvwxyz"])
        self.assertEqual(get_correct_id(lst, "a", True), "a-aa")

    def test_get_correct_path(self):
        self.assertEqual(get_correct_path(self.portal, "abcde"), "abcde")
        self.assertIn("folder", self.portal.objectIds())
        self.assertEqual(get_correct_path(self.portal, "folder"), "folder-1")
        self.assertEqual(get_correct_path(self.portal, "folder/abcde"), "folder/abcde")
        self.portal.folder.invokeFactory("Document", id="abcde", title="Document")
        self.assertEqual(get_correct_path(self.portal, "folder/abcde"), "folder/abcde-1")
        self.portal.folder.invokeFactory("Document", id="abcde-1", title="Document")
        self.assertEqual(get_correct_path(self.portal, "folder/abcde"), "folder/abcde-2")

    def test_filter_keys(self):
        dic = {"a": 1, "b": 2, "c": 3}
        self.assertListEqual(["a", "b", "c"], sorted(filter_keys(dic, []).keys()))
        self.assertListEqual([1, 2, 3], sorted(filter_keys(dic, []).values()))
        self.assertListEqual(["a", "b", "c"], sorted(filter_keys(dic, ["a", "b", "c"]).keys()))
        self.assertListEqual([1, 2, 3], sorted(filter_keys(dic, ["a", "b", "c"]).values()))
        self.assertListEqual(["a", "c"], sorted(filter_keys(dic, ["a", "c"]).keys()))
        self.assertListEqual([1, 3], sorted(filter_keys(dic, ["a", "c"]).values()))
        del dic["c"]
        self.assertListEqual(["a", "b", "c"], sorted(filter_keys(dic, ["a", "b", "c"]).keys()))
        values = filter_keys(dic, ["a", "b", "c"]).values()
        self.assertIn(None, values)
        self.assertListEqual([1, 2], sorted([v for v in values if v is not None]))  # in py3 sorted cannot handle None
        self.assertListEqual(
            ["", "1", "2"], sorted([str(v) for v in filter_keys(dic, ["a", "b", "c"], unfound="").values()])
        )

    def test_get_obj_from_path(self):
        folder = self.portal.folder
        self.assertEqual(get_obj_from_path(self.portal, path="/folder"), folder)
        self.assertEqual(get_obj_from_path(self.portal, path="folder"), folder)
        self.assertEqual(get_obj_from_path(self.portal, path=u"folder"), folder)
        self.assertEqual(get_obj_from_path(self.portal, {"key": "folder"}, path_key="key"), folder)
        self.assertIsNone(get_obj_from_path(self.portal, {}, path_key="unknown_key"))
        self.assertIsNone(get_obj_from_path(self.portal, path="in_va_lid"))

    def test_get_main_path(self):
        orig_home = os.getenv("INSTANCE_HOME")
        orig_pwd = os.getenv("PWD")
        self.assertEqual(get_main_path("/etc"), "/etc")
        os.environ["INSTANCE_HOME"] = "/etc/parts/xx"
        self.assertEqual(get_main_path(), "/etc")
        os.environ["INSTANCE_HOME"] = "/etc/abcd"
        os.environ["PWD"] = "/myhome"
        self.assertRaisesRegex(Exception, u"Path '/myhome' doesn't exist", get_main_path)  # path doesn't exist
        self.assertRaisesRegex(Exception, u"Path '/myhome/toto' doesn't exist", get_main_path, subpath="toto")
        self.assertEqual(get_main_path("/", "etc"), "/etc")
        os.environ["INSTANCE_HOME"] = orig_home
        os.environ["PWD"] = orig_pwd

    def test_key_val(self):
        dic = {"key": "val"}
        self.assertEqual(key_val("key", dic), "val")
        self.assertEqual(key_val("unknown", dic), "unknown")

    def test_pool_tuples(self):
        lst = [1, 2, 3, 4, 5, 6]
        self.assertEqual(pool_tuples(None, 1), None)
        self.assertListEqual(pool_tuples([], 1), [])
        self.assertListEqual([t for t in pool_tuples(lst, 1)], [(1,), (2,), (3,), (4,), (5,), (6,)])
        self.assertListEqual([t for t in pool_tuples(lst, 2)], [(1, 2), (3, 4), (5, 6)])
        self.assertListEqual([t for t in pool_tuples(lst, 3)], [(1, 2, 3), (4, 5, 6)])
        self.assertListEqual([t for t in pool_tuples(lst, 6)], [(1, 2, 3, 4, 5, 6)])
        with self.assertRaises(Exception) as cm:
            pool_tuples([1, 2, 3], 2, e_msg=u"My pool")
        self.assertEqual(
            str(cm.exception),
            "My pool: the given iterable must contain a multiple of 2 elements: " "value = '[1, 2, 3]'",
        )

    def test_relative_path(self):
        self.assertEqual(relative_path(self.portal, "/plone/directory"), "/directory")
        self.assertEqual(relative_path(self.portal, "/plone/directory", False), "directory")
        self.assertEqual(relative_path(self.portal, "/alone/directory"), "/alone/directory")

    def test_split_text(self):
        self.assertTupleEqual(split_text(None, 50), (u"", u""))
        self.assertTupleEqual(split_text("aa bb cc dd éé", 50), (u"aa bb cc dd éé", u""))
        self.assertTupleEqual(split_text(u"aa bb cc dd ee", 50), (u"aa bb cc dd ee", u""))
        self.assertTupleEqual(split_text(u"aa bb cc dd ee", 0), (u"", u"aa bb cc dd ee"))
        self.assertTupleEqual(split_text(u"aa bb cc dd ee", 7), (u"aa bb ", u"cc dd ee"))
        self.assertTupleEqual(split_text(u"aa bb cc dd ee", 8), (u"aa bb cc", u" dd ee"))
        self.assertTupleEqual(split_text(u"aa bb cc dd ee", 9), (u"aa bb cc ", u"dd ee"))
        # space before the middle
        self.assertTupleEqual(split_text(u"aa bbccddee", 9), (u"aa bbccdd", u"ee"))

    def test_str_to_bool(self):
        dic = {
            "True": "True",
            "true": "true",
            "False": "False",
            "false": "false",
            "0": "0",
            0: 0,
            "1": "1",
            123: 123,
            "a": "a",
            "errors": 0,
        }
        self.assertTrue(str_to_bool(dic, "True", logger))
        self.assertTrue(str_to_bool(dic, "true", logger))
        self.assertFalse(str_to_bool(dic, "False", logger))
        self.assertFalse(str_to_bool(dic, "false", logger))
        self.assertFalse(str_to_bool(dic, "0", logger))  # using fake log method to test error
        self.assertFalse(str_to_bool(dic, 0, logger))
        self.assertTrue(str_to_bool(dic, "1", logger))
        self.assertTrue(str_to_bool(dic, 123, logger))
        self.assertEqual(dic["errors"], 0)
        self.assertFalse(str_to_bool(dic, "a", logger))
        self.assertEqual(dic["errors"], 1)

    def test_str_to_date(self):
        dic = {1: "2023/03/03", 2: "2023/03/03 09:29", "errors": 0}
        # str_to_date(item, key, log_method, fmt='%Y/%m/%d', can_be_empty=True, as_date=True, **log_params):
        self.assertIsNone(str_to_date(dic, "bad_key", logger))
        self.assertRaises(TypeError, str_to_date, dic, "bad_key", logger, can_be_empty=False)
        self.assertIsNone(str_to_date(dic, 1, logger, min_val=datetime.date(2023, 4, 1)))
        self.assertEqual(dic["errors"], 1)
        self.assertIsNone(str_to_date(dic, 1, logger, max_val=datetime.date(2022, 4, 1)))
        self.assertEqual(dic["errors"], 2)
        self.assertIsNone(
            str_to_date(dic, 1, logger, min_val=datetime.date(2023, 4, 1), max_val=datetime.date(2022, 4, 1))
        )
        self.assertEqual(dic["errors"], 3)
        ret = str_to_date(dic, 1, logger)
        self.assertIsInstance(ret, datetime.date)
        self.assertEqual(str(ret), "2023-03-03")
        ret = str_to_date(dic, 2, logger, fmt="%Y/%m/%d %H:%M", as_date=False)
        self.assertIsInstance(ret, datetime.datetime)
        self.assertEqual(str(ret), "2023-03-03 09:29:00")
        ret = str_to_date(dic, 2, logger, fmt="%Y/%m/%d %M:%H", as_date=False)
        self.assertEqual(dic["errors"], 4)



# -*- coding: utf-8 -*-

from imio.helpers import barcode

import os
import unittest


class TestBarcode(unittest.TestCase):
    def test_generate_barcode(self):
        current_dir = os.path.dirname(os.path.abspath(__file__))
        barcode_resource = "barcode_python3_zint_60x60_156bytes.png"
        with open(os.path.join(current_dir, barcode_resource), "rb") as barcode_file:
            bc = barcode_file.read()
            result = barcode.generate_barcode("123")
            result = result.read()
            self.assertEqual(bc, result)

    def test_generate_barcode_filetype(self):
        result = barcode.generate_barcode("123", filetype="GIF")
        self.assertTrue(result.read().startswith(b"GIF"))

    def test_generate_barcode_missing_executable(self):
        self.assertRaises(OSError, barcode.generate_barcode, "123", executable="zints")



# -*- coding: utf-8 -*-

from imio.helpers import get_cachekey_volatile
from imio.helpers.content import get_user_fullname
from natsort import humansorted
from operator import attrgetter
from plone import api
from plone.memoize import ram
from Products.CMFPlone.utils import safe_unicode
from zope.i18n import translate
from zope.interface import implementer
from zope.schema.interfaces import IVocabularyFactory
from zope.schema.vocabulary import SimpleTerm
from zope.schema.vocabulary import SimpleVocabulary

import six


try:
    from plone.app.vocabularies.principals import UsersFactory  # Plone6
except ImportError:
    from plone.app.vocabularies.users import UsersFactory


def voc_cache_key(method, self, context=None, query=""):
    """Returns a persistent portal stored date following the given cache key.

    Must be programatically invalidated."""
    return get_cachekey_volatile("_users_groups_value")


def get_users_voc(with_userid):
    acl_users = api.portal.get_tool("acl_users")
    users = acl_users.searchUsers(sort_by="")
    terms = []
    # manage duplicates, this can be the case when using LDAP and same userid in source_users
    userids = []
    for user in users:
        user_id = user["id"]
        if user_id not in userids:
            userids.append(user_id)
            # bypass special characters, may happen when using LDAP
            try:
                six.text_type(user_id)
            except UnicodeDecodeError:
                continue
            if with_userid:
                term_title = u"{0} ({1})".format(get_user_fullname(user_id), user_id)
            else:
                term_title = get_user_fullname(user_id)
            term = SimpleTerm(user_id, user_id, term_title)
            terms.append(term)
    terms = humansorted(terms, key=attrgetter("title"))
    return SimpleVocabulary(terms)


class SortedUsers(UsersFactory):
    """Create a users vocabulary with userid as value.
    Append ' (userid)' to term title."""

    @ram.cache(voc_cache_key)
    def SortedUsers__call__(self, context=None, query=""):
        return get_users_voc(True)

    __call__ = SortedUsers__call__


SortedUsersFactory = SortedUsers()


class SimplySortedUsers(SortedUsers):
    """With userid as value and fullname as title"""

    @ram.cache(voc_cache_key)
    def SimplySortedUsers__call__(self, context=None, query=""):
        return get_users_voc(False)

    __call__ = SimplySortedUsers__call__


SimplySortedUsersFactory = SimplySortedUsers()


class EnhancedTerm(SimpleTerm):
    def __init__(self, value, token=None, title=None, **attrs):
        super(EnhancedTerm, self).__init__(value, token=token, title=title)
        self.attrs = attrs


@implementer(IVocabularyFactory)
class YesNoForFacetedVocabulary(object):
    """Vocabulary to be used with a faceted radio widget on a field index containing prefixed 0 and 1 strings."""

    prefix = ""

    def __call__(self, context):
        """ """
        res = [
            SimpleTerm(
                self.prefix + "0",
                self.prefix + "0",
                safe_unicode(translate("yesno_value_false", domain="imio.helpers", context=context.REQUEST)),
            ),
            SimpleTerm(
                self.prefix + "1",
                self.prefix + "1",
                safe_unicode(translate("yesno_value_true", domain="imio.helpers", context=context.REQUEST)),
            ),
        ]
        return SimpleVocabulary(res)
